from typing import Dict, List, Optional, Tuple
from promptmodel.database.models import (
    DeployedFunctionModel,
    DeployedFunctionModelVersion,
    DeployedPrompt,
)
from playhouse.shortcuts import model_to_dict
from promptmodel.utils.random_utils import select_version_by_ratio
from promptmodel.utils import logger
from promptmodel.database.config import db
from promptmodel.database.crud_chat import *


# Insert

# Select all

# Select one


def get_deployed_prompts(function_model_name: str) -> Tuple[List[DeployedPrompt], str]:
    try:
        with db.atomic():
            versions: List[DeployedFunctionModelVersion] = list(
                DeployedFunctionModelVersion.select()
                .join(DeployedFunctionModel)
                .where(
                    DeployedFunctionModelVersion.function_model_uuid
                    == DeployedFunctionModel.get(
                        DeployedFunctionModel.name == function_model_name
                    ).uuid
                )
            )
            prompts: List[DeployedPrompt] = list(
                DeployedPrompt.select()
                .where(
                    DeployedPrompt.version_uuid.in_(
                        [version.uuid for version in versions]
                    )
                )
                .order_by(DeployedPrompt.step.asc())
            )
        # select version by ratio
        selected_version = select_version_by_ratio(
            [version.__data__ for version in versions]
        )
        selected_prompts = list(
            filter(
                lambda prompt: str(prompt.version_uuid.uuid)
                == str(selected_version["uuid"]),
                prompts,
            )
        )

        version_details = {
            "model": selected_version["model"],
            "version" : selected_version["version"],
            "uuid": selected_version["uuid"],
            "parsing_type": selected_version["parsing_type"],
            "output_keys": selected_version["output_keys"],
        }

        return selected_prompts, version_details
    except Exception as e:
        logger.error(e)
        return None, None


# Update


async def update_deployed_cache(project_status: dict):
    """Update Deployed Prompts Cache"""
    # TODO: 효율적으로 수정
    # 현재는 delete all & insert all
    function_models = project_status["function_models"]
    function_model_versions = project_status["function_model_versions"]
    for version in function_model_versions:
        if version["is_published"] is True:
            version["ratio"] = 1.0
    prompts = project_status["prompts"]

    with db.atomic():
        DeployedFunctionModel.delete().execute()
        DeployedFunctionModelVersion.delete().execute()
        DeployedPrompt.delete().execute()
        DeployedFunctionModel.insert_many(function_models).execute()
        DeployedFunctionModelVersion.insert_many(function_model_versions).execute()
        DeployedPrompt.insert_many(prompts).execute()
    return



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

import nest_asyncio
from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from ..constants import function_shemas
from promptmodel.llms.llm import LLM
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import *
from promptmodel.types.enums import ParsingType
from promptmodel.utils.async_utils import run_async_in_sync

html_output_format = """\
You must follow the provided output format. Keep the string between <> as it is.
Output format:
<response type=str>
(value here)
</response>
"""

double_bracket_output_format = """\
You must follow the provided output format. Keep the string between [[ ]] as it is.
Output format:
[[response type=str]]
(value here)
[[/response]]
"""

colon_output_format = """\
You must follow the provided output format. Keep the string before : as it is
Output format:
response type=str: (value here)

"""

tool_schemas = [{"type": "function", "function": schema} for schema in function_shemas]


def test_stream_with_tools(mocker):
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    stream_res: Generator[LLMStreamResponse, None, None] = llm.stream(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
    )
    print("HELLO")
    error_count = 0
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    chunks: List[LLMStreamResponse] = []
    for res in stream_res:
        print(res)
        chunks.append(res)
        if res.error:
            error_count += 1
        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)
            final_response = res
    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"
    assert (
        getattr(final_response.api_response.choices[0].message, "tool_calls", None)
        is not None
    ), "tool_calls is None"
    assert isinstance(
        final_response.api_response.choices[0].message.tool_calls[0],
        ChatCompletionMessageToolCall,
    )
    assert len([c for c in chunks if c.tool_calls is not None]) > 0
    assert isinstance(chunks[1].tool_calls[0], ChoiceDeltaToolCall)

    assert api_responses[0].choices[0].message.content is None, "content is not None"
    assert api_responses[0]._response_ms is not None, "response_ms is None"
    assert api_responses[0].usage is not None, "usage is None"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    nest_asyncio.apply()
    run_async_in_sync(
        llm_proxy._async_log_to_cloud(
            log_uuid="test",
            version_uuid="test",
            inputs={},
            api_response=api_responses[0],
            parsed_outputs={},
            metadata={},
        )
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args
    api_response_dict = api_responses[0].model_dump()
    api_response_dict.update({"response_ms": api_responses[0]._response_ms})

    assert (
        kwargs["json"]["api_response"] == api_response_dict
    ), "api_response is not equal"


@pytest.mark.asyncio
async def test_astream_with_tools(mocker):
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    stream_res: AsyncGenerator[LLMStreamResponse, None] = llm.astream(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
    )

    error_count = 0
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    chunks: List[LLMStreamResponse] = []
    async for res in stream_res:
        chunks.append(res)
        if res.error:
            error_count += 1
        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)
            final_response = res

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"
    assert (
        getattr(final_response.api_response.choices[0].message, "tool_calls", None)
        is not None
    ), "tool_calls is None"
    assert isinstance(
        final_response.api_response.choices[0].message.tool_calls[0],
        ChatCompletionMessageToolCall,
    )
    assert len([c for c in chunks if c.tool_calls is not None]) > 0
    assert isinstance(chunks[1].tool_calls[0], ChoiceDeltaToolCall)

    assert api_responses[0].choices[0].message.content is None, "content is not None"
    assert api_responses[0]._response_ms is not None, "response_ms is None"
    assert api_responses[0].usage is not None, "usage is None"
    # assert api_responses[0].usage.prompt_tokens == 74, "prompt_tokens is not 74"

    # test logging
    llm_proxy = LLMProxy("test")

    mock_execute = AsyncMock()
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_execute.return_value = mock_response
    mocker.patch("promptmodel.llms.llm_proxy.AsyncAPIClient.execute", new=mock_execute)

    nest_asyncio.apply()
    await llm_proxy._async_log_to_cloud(
        log_uuid="test",
        version_uuid="test",
        inputs={},
        api_response=api_responses[0],
        parsed_outputs={},
        metadata={},
    )

    mock_execute.assert_called_once()
    _, kwargs = mock_execute.call_args
    api_response_dict = api_responses[0].model_dump()
    api_response_dict.update({"response_ms": api_responses[0]._response_ms})

    assert (
        kwargs["json"]["api_response"] == api_response_dict
    ), "api_response is not equal"


def test_stream_and_parse_with_tools(mocker):
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    stream_res: Generator[LLMStreamResponse, None, None] = llm.stream_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=None,
    )

    error_count = 0
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    chunks: List[LLMStreamResponse] = []
    for res in stream_res:
        chunks.append(res)
        if res.error:
            error_count += 1
            print("ERROR")
            print(res.error)
            print(res.error_log)
        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)
            final_response = res

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"
    assert (
        getattr(final_response.api_response.choices[0].message, "tool_calls", None)
        is not None
    ), "tool_calls is None"
    assert isinstance(
        final_response.api_response.choices[0].message.tool_calls[0],
        ChatCompletionMessageToolCall,
    )
    assert len([c for c in chunks if c.tool_calls is not None]) > 0
    assert isinstance(chunks[1].tool_calls[0], ChoiceDeltaToolCall)

    assert api_responses[0].choices[0].message.content is None, "content is not None"

    # Not call function, parsing case
    messages = [
        {
            "role": "user",
            "content": "Hello, How are you?\n" + html_output_format,
        }
    ]
    stream_res: Generator[LLMStreamResponse, None, None] = llm.stream_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )

    error_count = 0
    error_log = ""
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    chunks: List[LLMStreamResponse] = []
    for res in stream_res:
        chunks.append(res)
        if res.error:
            error_count += 1
            error_log = res.error_log
            print("ERROR")
        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)
            final_response = res

    if not "str" in api_responses[0].choices[0].message.content:
        # if "str" in content, just LLM make mistake in generation.
        assert (
            error_count == 0
        ), f"error_count is not 0, {error_log}, {api_responses[0].model_dump()}"
        assert final_response.parsed_outputs is not None, "parsed_outputs is None"
    assert len(api_responses) == 1, "api_count is not 1"
    assert (
        getattr(final_response.api_response.choices[0].message, "tool_calls", None)
        is None
    ), "tool_calls is not None"
    assert len([c for c in chunks if c.tool_calls is not None]) == 0

    assert api_responses[0].choices[0].message.content is not None, "content is None"


@pytest.mark.asyncio
async def test_astream_and_parse_with_tools(mocker):
    messages = [{"role": "user", "content": "What is the weather like in Boston?"}]
    llm = LLM()
    stream_res: AsyncGenerator[LLMStreamResponse, None] = llm.astream_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=None,
    )

    error_count = 0
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    chunks: List[LLMStreamResponse] = []
    async for res in stream_res:
        chunks.append(res)
        if res.error:
            error_count += 1

        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)
            final_response = res

    assert error_count == 0, "error_count is not 0"
    assert len(api_responses) == 1, "api_count is not 1"

    assert api_responses[0].choices[0].message.content is None, "content is not None"
    assert (
        getattr(final_response.api_response.choices[0].message, "tool_calls", None)
        is not None
    ), "tool_calls is None"
    assert isinstance(
        final_response.api_response.choices[0].message.tool_calls[0],
        ChatCompletionMessageToolCall,
    )
    assert len([c for c in chunks if c.tool_calls is not None]) > 0
    assert isinstance(chunks[1].tool_calls[0], ChoiceDeltaToolCall)

    # Not call function, parsing case
    messages = [
        {
            "role": "user",
            "content": "Hello, How are you?\n" + html_output_format,
        }
    ]
    stream_res: AsyncGenerator[LLMStreamResponse, None] = llm.astream_and_parse(
        messages=messages,
        tools=tool_schemas,
        model="gpt-4-1106-preview",
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )

    error_count = 0
    error_log = ""
    api_responses: List[ModelResponse] = []
    final_response: Optional[LLMStreamResponse] = None
    chunks: List[LLMStreamResponse] = []
    async for res in stream_res:
        chunks.append(res)
        if res.error:
            error_count += 1
            error_log = res.error_log

        if (
            res.api_response
            and getattr(res.api_response.choices[0], "delta", None) is None
        ):
            api_responses.append(res.api_response)
            final_response = res

    if not "str" in api_responses[0].choices[0].message.content:
        # if "str" in content, just LLM make mistake in generation.
        assert (
            error_count == 0
        ), f"error_count is not 0, {error_log}, {api_responses[0].model_dump()}"
        assert final_response.parsed_outputs is not None, "parsed_outputs is None"
    assert len(api_responses) == 1, "api_count is not 1"
    assert (
        getattr(final_response.api_response.choices[0].message, "tool_calls", None)
        is None
    ), "tool_calls is not None"

    assert len([c for c in chunks if c.tool_calls is not None]) == 0

    assert api_responses[0].choices[0].message.content is not None, "content is None"



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from promptmodel.llms.llm import LLM, ParseResult
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import LLMResponse, LLMStreamResponse
from promptmodel.types.enums import ParsingType

proxy = LLMProxy(name="test")


def test_run(mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: LLMResponse = proxy.run({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None


@pytest.mark.asyncio
async def test_arun(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: LLMResponse = await proxy.arun({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None


def test_stream(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: Generator[LLMStreamResponse, None, None] = proxy.stream({})
    chunks: List[LLMStreamResponse] = []
    for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0


@pytest.mark.asyncio
async def test_astream(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: AsyncGenerator[LLMStreamResponse, None] = proxy.astream({})
    chunks: List[LLMStreamResponse] = []
    async for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0


def test_run_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: LLMResponse = proxy.run({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()

    # mock run
    mock_run = MagicMock()
    mock_run.return_value = LLMResponse(
        api_response=res.api_response, parsed_outputs={"key": "value"}
    )
    mocker.patch("promptmodel.llms.llm.LLM.run", mock_run)
    mock_res = proxy.run({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert mock_res.raw_output is None
    assert mock_res.error is None or res.error is False
    assert mock_res.api_response is not None
    assert mock_res.parsed_outputs is not None


@pytest.mark.asyncio
async def test_arun_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: LLMResponse = await proxy.arun({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()

    # mock run
    mock_run = AsyncMock()
    mock_run.return_value = LLMResponse(
        api_response=res.api_response, parsed_outputs={"key": "value"}
    )
    mocker.patch("promptmodel.llms.llm.LLM.arun", mock_run)
    mock_res: LLMResponse = await proxy.arun({})
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()
    assert mock_res.raw_output is None
    assert mock_res.error is None or res.error is False
    assert mock_res.api_response is not None
    assert mock_res.parsed_outputs is not None


def test_stream_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: Generator[LLMStreamResponse, None, None] = proxy.stream({})
    chunks: List[LLMStreamResponse] = []
    for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()

    def mock_stream_generator(*args, **kwargs):
        yield LLMStreamResponse(parsed_outputs={"key": "value"})
        yield LLMStreamResponse(api_response=chunks[-1].api_response)

    mock_run = MagicMock(side_effect=mock_stream_generator)
    mocker.patch("promptmodel.llms.llm.LLM.stream", mock_run)

    mock_res: Generator[LLMStreamResponse, None, None] = proxy.stream({})
    mock_chunks: List[LLMStreamResponse] = []
    for chunk in mock_res:
        mock_chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert mock_chunks[-1].api_response is not None
    assert len([chunk for chunk in mock_chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in mock_chunks if chunk.parsed_outputs is not None]) > 0
    assert len([chunk for chunk in mock_chunks if chunk.raw_output is not None]) == 0


@pytest.mark.asyncio
async def test_astream_and_parse(
    mocker, mock_fetch_prompts: AsyncMock, mock_async_log_to_cloud: AsyncMock
):
    fetch_prompts = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_prompts", mock_fetch_prompts
    )
    async_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_log_to_cloud",
        mock_async_log_to_cloud,
    )

    res: AsyncGenerator[LLMStreamResponse, None] = proxy.astream({})
    chunks: List[LLMStreamResponse] = []
    async for chunk in res:
        chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0

    fetch_prompts.reset_mock()
    async_log_to_cloud.reset_mock()

    async def mock_stream_generator(*args, **kwargs):
        yield LLMStreamResponse(parsed_outputs={"key": "value"})
        yield LLMStreamResponse(api_response=chunks[-1].api_response)

    mock_run = MagicMock(side_effect=mock_stream_generator)
    mocker.patch("promptmodel.llms.llm.LLM.astream", mock_run)

    mock_res: AsyncGenerator[LLMStreamResponse, None] = proxy.astream({})
    mock_chunks: List[LLMStreamResponse] = []
    async for chunk in mock_res:
        mock_chunks.append(chunk)
    fetch_prompts.assert_called_once()
    async_log_to_cloud.assert_called_once()

    assert mock_chunks[-1].api_response is not None
    assert len([chunk for chunk in mock_chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in mock_chunks if chunk.parsed_outputs is not None]) > 0
    assert len([chunk for chunk in mock_chunks if chunk.raw_output is not None]) == 0



"""This single file is needed to build the DevApp development dashboard."""

from promptmodel import DevApp

# Example imports
# from <dirname> import <objectname>
from main import dev as main_client

app = DevApp()

# Example usage
# This is needed to integrate your codebase with the prompt engineering dashboard
# app.include_client(<objectname>)
app.include_client(main_client)


import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from promptmodel.function_model import RegisteringMeta


def test_registering_meta(mocker):
    # Fail to find DevClient instance
    client_instance = RegisteringMeta.find_client_instance()
    assert client_instance is None



import pytest
from unittest.mock import AsyncMock, patch, MagicMock
import asyncio

from websockets.exceptions import ConnectionClosedOK
from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.types.enums import LocalTask


@pytest.mark.asyncio
async def test_connect_to_gateway(
    mocker, websocket_client: DevWebsocketClient, mock_websocket: AsyncMock
):
    project_uuid = "test_uuid"
    connection_name = "test_project"
    cli_access_header = {"Authorization": "Bearer testtoken"}

    with patch.object(
        websocket_client, "_DevWebsocketClient__handle_message", new_callable=AsyncMock
    ) as mock_function:
        mock_function.side_effect = ConnectionClosedOK(None, None)
        with patch(
            "promptmodel.websocket.websocket_client.connect",
            new_callable=MagicMock,
            return_value=mock_websocket,
        ) as mock_connect:
            # 5초 후에 자동으로 테스트 종료
            await websocket_client.connect_to_gateway(
                project_uuid, connection_name, cli_access_header, retries=1
            ),
            mock_connect.assert_called_once()
            mock_websocket.recv.assert_called_once()
            websocket_client._DevWebsocketClient__handle_message.assert_called_once()


@pytest.mark.asyncio
async def test_local_tasks(
    mocker, websocket_client: DevWebsocketClient, mock_websocket: AsyncMock
):
    websocket_client._devapp.function_models = {}
    websocket_client._devapp.samples = {}
    websocket_client._devapp.functions = {"test_function": "test_function"}

    await websocket_client._DevWebsocketClient__handle_message(
        message={"type": LocalTask.LIST_CODE_FUNCTIONS}, ws=mock_websocket
    )



from promptmodel import DevClient, FunctionModel
from promptmodel.types import FunctionModelConfig

dev = DevClient()

config: FunctionModelConfig = FunctionModel("new_prompt").get_config()

@dev.register
def summarize(text: str) -> str:
    res = FunctionModel("summarize").run({"text" : text})
    print(res.raw_output)
    
# main
if __name__ == "__main__":
    summarize("This is a test")


import pytest
from unittest.mock import AsyncMock, patch, MagicMock

import asyncio
from typing import Optional
from uuid import uuid4
from dataclasses import dataclass
from websockets.exceptions import ConnectionClosedOK
from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.types.enums import LocalTask, LocalTaskErrorType
from promptmodel.types.response import FunctionSchema


@dataclass
class ChatModelInterface:
    name: str
    default_model: str = "gpt-3.5-turbo"


def get_current_weather(location: str, unit: Optional[str] = "celsius"):
    return "13 degrees celsius"


get_current_weather_desc = {
    "name": "get_current_weather",
    "description": "Get the current weather in a given location",
    "parameters": {
        "type": "object",
        "properties": {
            "location": {
                "type": "string",
                "description": "The city and state, e.g. San Francisco, CA",
            },
            "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
        },
        "required": ["location"],
    },
}


@pytest.mark.asyncio
async def test_run_model_function_call(
    mocker,
    websocket_client: DevWebsocketClient,
    mock_websocket: AsyncMock,
    mock_json_dumps: MagicMock,
):
    websocket_client._devapp.chat_models = [ChatModelInterface("test_module")]
    websocket_client._devapp.samples = {
        "sample_1": {"user_message": "What is the weather like in Boston?"}
    }
    websocket_client._devapp.functions = {
        "get_current_weather": {
            "schema": FunctionSchema(**get_current_weather_desc),
            "function": get_current_weather,
        }
    }

    function_schemas_in_db = [FunctionSchema(**get_current_weather_desc).model_dump()]

    function_schemas_in_db[0]["mock_response"] = "13"

    mocker.patch("promptmodel.websocket.websocket_client.json.dumps", mock_json_dumps)

    # success case with function_call
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_CHAT_MODEL,
            "chat_model_name": "test_module",
            "system_prompt": {
                "role": "system",
                "content": "You are a helpful assistant.",
            },
            "old_messages": [],
            "new_messages": [
                {
                    "role": "user",
                    "content": "What is the weather in Boston?",
                }
            ],
            "model": "gpt-3.5-turbo",
            "functions": ["get_current_weather"],
            "function_schemas": function_schemas_in_db,
        },
        ws=mock_websocket,
    )

    call_args_list = mock_websocket.send.call_args_list
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 0
    assert len([d for d in data if d["status"] == "completed"]) == 1

    assert len([d for d in data if "function_response" in d]) == 1
    assert [d for d in data if "function_response" in d][0]["function_response"][
        "name"
    ] == "get_current_weather"

    assert len([d for d in data if "function_call" in d]) > 1
    assert len([d for d in data if "raw_output" in d]) > 0

    mock_websocket.send.reset_mock()
    function_schemas_in_db[0]["mock_response"] = "13"

    print(
        "======================================================================================="
    )

    # success case with no function call
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_CHAT_MODEL,
            "chat_model_name": "test_module",
            "system_prompt": {
                "role": "system",
                "content": "You are a helpful assistant.",
            },
            "old_messages": [],
            "new_messages": [
                {
                    "role": "user",
                    "content": "Hello?",
                }
            ],
            "model": "gpt-3.5-turbo",
        },
        ws=mock_websocket,
    )

    call_args_list = mock_websocket.send.call_args_list
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 0
    assert len([d for d in data if d["status"] == "completed"]) == 1

    assert len([d for d in data if "function_response" in d]) == 0

    assert len([d for d in data if "function_call" in d]) == 0
    assert len([d for d in data if "raw_output" in d]) > 0

    mock_websocket.send.reset_mock()
    function_schemas_in_db[0]["mock_response"] = "13"

    print(
        "======================================================================================="
    )

    # FUNCTION_CALL_FAILED_ERROR case
    def error_raise_function(*args, **kwargs):
        raise Exception("error")

    websocket_client._devapp.functions = {
        "get_current_weather": {
            "schema": FunctionSchema(**get_current_weather_desc),
            "function": error_raise_function,
        }
    }

    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_CHAT_MODEL,
            "chat_model_name": "test_module",
            "system_prompt": {
                "role": "system",
                "content": "You are a helpful assistant.",
            },
            "old_messages": [],
            "new_messages": [
                {
                    "role": "user",
                    "content": "What is the weather in Boston?",
                }
            ],
            "model": "gpt-3.5-turbo",
            "functions": ["get_current_weather"],
            "function_schemas": function_schemas_in_db,
        },
        ws=mock_websocket,
    )
    call_args_list = mock_websocket.send.call_args_list
    # print(call_args_list)
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 1
    assert [d for d in data if d["status"] == "failed"][0][
        "error_type"
    ] == LocalTaskErrorType.FUNCTION_CALL_FAILED_ERROR.value
    assert len([d for d in data if d["status"] == "completed"]) == 0
    assert len([d for d in data if "function_response" in d]) == 0
    assert len([d for d in data if "function_call" in d]) > 1
    assert len([d for d in data if "raw_output" in d]) == 0
    mock_websocket.send.reset_mock()
    function_schemas_in_db[0]["mock_response"] = "13"

    # function not in code case, should use mock_response
    websocket_client._devapp.functions = {}
    await websocket_client._DevWebsocketClient__handle_message(
        message={
            "type": LocalTask.RUN_CHAT_MODEL,
            "chat_model_name": "test_module",
            "system_prompt": {
                "role": "system",
                "content": "You are a helpful assistant.",
            },
            "old_messages": [],
            "new_messages": [
                {
                    "role": "user",
                    "content": "What is the weather in Boston?",
                }
            ],
            "model": "gpt-3.5-turbo",
            "functions": ["get_weather"],
            "function_schemas": function_schemas_in_db,
        },
        ws=mock_websocket,
    )
    call_args_list = mock_websocket.send.call_args_list
    print(call_args_list)
    data = [arg.args[0] for arg in call_args_list]

    assert len([d for d in data if d["status"] == "failed"]) == 0
    assert len([d for d in data if d["status"] == "completed"]) == 1

    assert len([d for d in data if "function_response" in d]) == 1
    assert (
        "FAKE RESPONSE"
        in [d for d in data if "function_response" in d][0]["function_response"][
            "response"
        ]
    )
    assert len([d for d in data if "function_call" in d]) > 1
    assert len([d for d in data if "raw_output" in d]) > 0



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from ..constants import function_shemas
from promptmodel.llms.llm import LLM, ParseResult
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import (
    LLMResponse,
    LLMStreamResponse,
    Message,
    Delta,
    Choices,
    Usage,
    FunctionCall,
    ChoiceDeltaFunctionCall,
    StreamingChoices,
)
from promptmodel.types.enums import ParsingType


def test_parse_output_pattern_error_cases(mocker):
    llm = LLM()
    res: ParseResult = llm.__parse_output_pattern__(parsing_type=None)
    assert res.parsed_outputs == {}
    assert res.error == False
    assert res.error_log is None

    res: ParseResult = llm.__parse_output_pattern__(
        parsing_type=ParsingType.COLON.value, raw_output=None
    )
    assert res.parsed_outputs == {}
    assert res.error == True
    assert res.error_log is not None

    res: ParseResult = llm.__parse_output_pattern__(
        parsing_type=ParsingType.HTML.value,
        raw_output="<response type=int>hello</response>",
    )
    assert res.parsed_outputs == {}
    assert res.error == True
    assert res.error_log is not None


def test_run_error_cases(mocker):
    llm = LLM()
    mocker.patch("promptmodel.llms.llm.completion", return_value={})
    res: LLMResponse = llm.run(
        messages=[{"role": "user", "content": "hello"}],
    )
    assert res.api_response is not None
    assert res.error == True

    mocker.patch("promptmodel.llms.llm.completion", side_effect=Exception("test"))
    res: LLMResponse = llm.run(
        messages=[{"role": "user", "content": "hello"}],
    )
    assert res.api_response is None
    assert res.error == True


@pytest.mark.asyncio
async def test_arun_error_cases(mocker):
    llm = LLM()
    mocker.patch("promptmodel.llms.llm.acompletion", return_value={})
    res: LLMResponse = await llm.arun(
        messages=[{"role": "user", "content": "hello"}],
    )
    assert res.api_response is not None
    assert res.error == True

    mocker.patch("promptmodel.llms.llm.acompletion", side_effect=Exception("test"))
    res: LLMResponse = await llm.arun(
        messages=[{"role": "user", "content": "hello"}],
    )
    assert res.api_response is None
    assert res.error == True


def test_stream_error_cases(mocker):
    llm = LLM()

    mocker.patch("promptmodel.llms.llm.completion", side_effect=Exception("test"))
    res: LLMResponse = llm.stream(
        messages=[{"role": "user", "content": "hello"}],
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
    assert len(results) == 1
    assert results[0].api_response is None
    assert results[0].error == True


@pytest.mark.asyncio
async def test_astream_error_cases(mocker):
    llm = LLM()

    mocker.patch("promptmodel.llms.llm.acompletion", side_effect=Exception("test"))
    res: LLMResponse = llm.astream(
        messages=[{"role": "user", "content": "hello"}],
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)

    assert len(results) == 1
    assert results[0].api_response is None
    assert results[0].error == True


def string_to_generator(input_string: str):
    def generator_format(response: str):
        model_response = ModelResponse(stream=True)
        model_response.choices[0] = StreamingChoices(
            **{
                "delta": Delta(**{"content": response}),
                "finish_reason": None,
            }
        )
        return model_response

    def generator():
        for char in input_string:
            yield generator_format(char)
        model_response = ModelResponse(stream=True)
        model_response.choices[0] = StreamingChoices(
            **{"delta": Delta(**{"content": None}), "finish_reason": "stop"}
        )
        yield model_response

    return generator()


async def string_to_agenerator(input_string: str):
    def generator_format(response: str):
        model_response = ModelResponse(stream=True)
        model_response.choices[0] = StreamingChoices(
            **{
                "delta": Delta(**{"content": response}),
                "finish_reason": None,
            }
        )
        return model_response

    async def agenerator():
        for char in input_string:
            yield generator_format(char)
        model_response = ModelResponse(stream=True)
        model_response.choices[0] = StreamingChoices(
            **{"delta": Delta(**{"content": None}), "finish_reason": "stop"}
        )
        yield model_response

    return agenerator()


def test_stream_and_parse_error_cases(mocker):
    llm = LLM()

    # error case - colon type
    res: LLMResponse = llm.stream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.COLON.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
    assert results[0].api_response is None
    assert results[0].error == True
    assert results[0].error_log == "Cannot stream colon type"

    # error case - side effect
    mocker.patch(
        "promptmodel.llms.llm.completion",
        side_effect=Exception("test"),
    )
    res: LLMResponse = llm.stream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0
    assert len([result for result in results if result.api_response is not None]) == 0

    # # html success case
    # mocker.patch(
    #     "promptmodel.llms.llm.completion",
    #     return_value=string_to_generator("<response type=str>hello</response>"),
    # )
    # res: LLMResponse = llm.stream(
    #     messages=[{"role": "user", "content": "hello"}],
    #     parsing_type=ParsingType.HTML.value,
    #     output_keys=["response"],
    # )
    # results: List[LLMStreamResponse] = []
    # for chunk in res:
    #     results.append(chunk)
    # assert len([result for result in results if result.error == True]) == 0

    # html failed case - error in parsing
    mocker.patch(
        "promptmodel.llms.llm.completion",
        return_value=string_to_generator("<response type=str>hello<response>"),
    )
    res: LLMResponse = llm.stream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0

    # html failed case - error in key matching
    mocker.patch(
        "promptmodel.llms.llm.completion",
        return_value=string_to_generator("<respond type=str>hello</respond>"),
    )
    res: LLMResponse = llm.stream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0
    error_log = [result.error_log for result in results if result.error == True][0]
    assert error_log == "Output keys do not match with parsed output keys"

    # html failed case - Having functions as input case
    mocker.patch(
        "promptmodel.llms.llm.completion",
        return_value=string_to_generator("<respond type=str>hello</respond>"),
    )
    res: LLMResponse = llm.stream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
        functions=function_shemas,
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
        # print(chunk.__dict__)
    assert len([result for result in results if result.error == True]) > 0
    error_log = [result.error_log for result in results if result.error == True][0]
    print([result.error_log for result in results if result.error == True])
    assert error_log == "Output keys do not match with parsed output keys"

    # Double Bracket failed case - error in parsing
    mocker.patch(
        "promptmodel.llms.llm.completion",
        return_value=string_to_generator("[[response type=str]]hello[[response]]"),
    )
    res: LLMResponse = llm.stream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.DOUBLE_SQUARE_BRACKET.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0

    # Double Bracket failed case - error in key matching
    mocker.patch(
        "promptmodel.llms.llm.completion",
        return_value=string_to_generator("[[respond type=str]]hello[[/respond]]"),
    )
    res: LLMResponse = llm.stream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.DOUBLE_SQUARE_BRACKET.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0
    error_log = [result.error_log for result in results if result.error == True][0]
    assert error_log == "Output keys do not match with parsed output keys"


@pytest.mark.asyncio
async def test_astream_and_parse_error_cases(mocker):
    llm = LLM()

    # error case - colon type
    res: AsyncGenerator[LLMResponse, None] = llm.astream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.COLON.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)
    assert results[0].api_response is None
    assert results[0].error == True
    assert results[0].error_log == "Cannot stream colon type"

    # error case - side effect
    mocker.patch(
        "promptmodel.llms.llm.acompletion",
        side_effect=Exception("test"),
    )
    res: AsyncGenerator[LLMResponse, None] = llm.astream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0
    assert len([result for result in results if result.api_response is not None]) == 0

    # # html success case
    # mocker.patch(
    #     "promptmodel.llms.llm.completion",
    #     return_value=string_to_generator("<response type=str>hello</response>"),
    # )
    # res: LLMResponse = llm.stream(
    #     messages=[{"role": "user", "content": "hello"}],
    #     parsing_type=ParsingType.HTML.value,
    #     output_keys=["response"],
    # )
    # results: List[LLMStreamResponse] = []
    # for chunk in res:
    #     results.append(chunk)
    # assert len([result for result in results if result.error == True]) == 0

    # html failed case - error in parsing
    mocker.patch(
        "promptmodel.llms.llm.acompletion",
        return_value=await string_to_agenerator("<response type=str>hello<response>"),
    )
    res: AsyncGenerator[LLMResponse, None] = llm.astream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0

    # html failed case - error in key matching
    mocker.patch(
        "promptmodel.llms.llm.acompletion",
        return_value=await string_to_agenerator("<respond type=str>hello</respond>"),
    )
    res: AsyncGenerator[LLMResponse, None] = llm.astream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0
    error_log = [result.error_log for result in results if result.error == True][0]
    assert error_log == "Output keys do not match with parsed output keys"

    # html failed case - Having functions as input case
    mocker.patch(
        "promptmodel.llms.llm.completion",
        return_value=await string_to_agenerator("<respond type=str>hello</respond>"),
    )
    res: LLMResponse = llm.astream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.HTML.value,
        output_keys=["response"],
        functions=function_shemas,
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0
    error_log = [result.error_log for result in results if result.error == True][0]
    assert error_log == "Output keys do not match with parsed output keys"

    # Double Bracket failed case - error in parsing
    mocker.patch(
        "promptmodel.llms.llm.acompletion",
        return_value=await string_to_agenerator(
            "[[response type=str]]hello[[response]]"
        ),
    )
    res: AsyncGenerator[LLMResponse, None] = llm.astream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.DOUBLE_SQUARE_BRACKET.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0

    # Double Bracket failed case - error in key matching
    mocker.patch(
        "promptmodel.llms.llm.acompletion",
        return_value=await string_to_agenerator(
            "[[respond type=str]]hello[[/respond]]"
        ),
    )
    res: AsyncGenerator[LLMResponse, None] = llm.astream_and_parse(
        messages=[{"role": "user", "content": "hello"}],
        parsing_type=ParsingType.DOUBLE_SQUARE_BRACKET.value,
        output_keys=["response"],
    )
    results: List[LLMStreamResponse] = []
    async for chunk in res:
        results.append(chunk)
    assert len([result for result in results if result.error == True]) > 0
    error_log = [result.error_log for result in results if result.error == True][0]
    assert error_log == "Output keys do not match with parsed output keys"



"""Base module for interacting with LLM APIs."""
import re
import json
import datetime
from typing import Any, AsyncGenerator, List, Dict, Optional, Generator
from attr import dataclass

from pydantic import BaseModel
from dotenv import load_dotenv
from litellm import completion, acompletion

from promptmodel.types.response import (
    LLMResponse,
    LLMStreamResponse,
    ModelResponse,
    Usage,
    Choices,
    Message,
    FunctionCall,
    ChatCompletionMessageToolCall,
    Function,
    ChoiceDeltaToolCallFunction,
    ChoiceDeltaToolCall,
)
from promptmodel.utils import logger
from promptmodel.types.enums import ParsingType, ParsingPattern, get_pattern_by_type
from promptmodel.utils.output_utils import convert_str_to_type, update_dict
from promptmodel.utils.token_counting import (
    num_tokens_for_messages,
    num_tokens_from_function_call_output,
    num_tokens_from_functions_input,
)

load_dotenv()


class Role:
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"
    FUNCTION = "function"


class OpenAIMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = ""
    function_call: Optional[Dict[str, Any]] = None
    name: Optional[str] = None


DEFAULT_MODEL = "gpt-3.5-turbo"


@dataclass
class ParseResult:
    parsed_outputs: Dict[str, Any]
    error: bool
    error_log: Optional[str]


class LLM:
    def __init__(self):
        pass

    @classmethod
    def __parse_output_pattern__(
        cls,
        raw_output: Optional[str] = None,
        parsing_type: Optional[ParsingType] = None,
    ) -> ParseResult:
        if parsing_type is None:
            return ParseResult(parsed_outputs={}, error=False, error_log=None)
        if raw_output is None:
            return ParseResult(parsed_outputs={}, error=True, error_log="No content")
        parsing_pattern = get_pattern_by_type(parsing_type)
        whole_pattern = parsing_pattern["whole"]
        parsed_results = re.findall(whole_pattern, raw_output, flags=re.DOTALL)
        parsed_outputs = {}
        error: bool = False
        error_log: str = None

        try:
            for parsed_result in parsed_results:
                key = parsed_result[0]
                type_str = parsed_result[1]
                value = convert_str_to_type(parsed_result[2], type_str)
                parsed_outputs[key] = value
        except Exception as e:
            error = True
            error_log = str(e)

        return ParseResult(
            parsed_outputs=parsed_outputs,
            error=error,
            error_log=error_log,
        )

    def __validate_openai_messages(
        self, messages: List[Dict[str, str]]
    ) -> List[OpenAIMessage]:
        """Validate and convert list of dictionaries to list of OpenAIMessage."""
        res = []
        for message in messages:
            res.append(OpenAIMessage(**message))
        return res

    def run(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Return the response from openai chat completion."""
        response = None
        if functions == []:
            functions = None
        try:
            response: ModelResponse = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
                tools=tools,
                api_key=api_key,
            )

            content: Optional[str] = getattr(
                response.choices[0].message, "content", None
            )

            call_func: Optional[FunctionCall] = getattr(
                response.choices[0].message, "function_call", None
            )

            call_tools: Optional[List[ChatCompletionMessageToolCall]] = getattr(
                response.choices[0].message, "tool_calls", None
            )

            return LLMResponse(
                api_response=response,
                raw_output=content,
                function_call=call_func if call_func else None,
                tool_calls=call_tools if call_tools else None,
            )
        except Exception as e:
            if response is not None:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    async def arun(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Return the response from openai chat completion."""
        if functions == []:
            functions = None
        response = None
        try:
            response: ModelResponse = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
                tools=tools,
                api_key=api_key,
            )
            content: Optional[str] = getattr(
                response.choices[0].message, "content", None
            )

            call_func: Optional[FunctionCall] = getattr(
                response.choices[0].message, "function_call", None
            )

            call_tools: Optional[ChatCompletionMessageToolCall] = getattr(
                response.choices[0].message, "tool_calls", None
            )

            return LLMResponse(
                api_response=response,
                raw_output=content,
                function_call=call_func if call_func else None,
                tool_calls=call_tools if call_tools else None,
            )

        except Exception as e:
            if response is not None:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    def stream(
        self,
        messages: List[Dict[str, str]],  # input
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        *args,
        **kwargs,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Stream openai chat completion."""
        if functions == []:
            functions = None
        response = None
        try:
            # load_prompt()
            start_time = datetime.datetime.now()
            response = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
                tools=tools,
                api_key=api_key,
            )

            for chunk in self.__llm_stream_response_generator__(
                messages, response, start_time, functions, tools
            ):
                yield chunk
        except Exception as e:
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def astream(
        self,
        messages: List[Dict[str, str]],
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        *args,
        **kwargs,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        """Parse & stream output from openai chat completion."""
        if functions == []:
            functions = None
        response = None
        try:
            start_time = datetime.datetime.now()
            response = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
                tools=tools,
                api_key=api_key,
            )

            async for chunk in self.__llm_stream_response_agenerator__(
                messages, response, start_time, functions, tools
            ):
                yield chunk
        except Exception as e:
            yield LLMStreamResponse(error=True, error_log=str(e))

    def run_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        """Parse and return output from openai chat completion."""
        if functions == []:
            functions = None
        response = None
        parsed_success = True
        parse_result = None
        error_log = None
        try:
            response: ModelResponse = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
                tools=tools,
                api_key=api_key,
            )
            raw_output = getattr(response.choices[0].message, "content", None)

            call_func: Optional[FunctionCall] = getattr(
                response.choices[0].message, "function_call", None
            )

            call_tools: Optional[List[ChatCompletionMessageToolCall]] = getattr(
                response.choices[0].message, "tool_calls", None
            )

            if not call_func and not call_tools:
                # function call does not appear in output

                parse_result: ParseResult = self.__parse_output_pattern__(
                    raw_output, parsing_type
                )

                # if output_keys exist & parsed_outputs does not match with output_keys -> error
                # if parse_result.error -> error
                if (
                    output_keys is not None
                    and set(parse_result.parsed_outputs.keys()) != set(output_keys)
                ) or parse_result.error:
                    parsed_success = False
                    error_log = (
                        "Output keys do not match with parsed output keys"
                        if not parse_result.error_log
                        else parse_result.error_log
                    )

            return LLMResponse(
                api_response=response,
                raw_output=raw_output,
                parsed_outputs=parse_result.parsed_outputs if parse_result else None,
                function_call=call_func if call_func else None,
                tool_calls=call_tools if call_tools else None,
                error=not parsed_success,
                error_log=error_log,
            )
        except Exception as e:
            if response is not None:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    async def arun_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
    ) -> LLMResponse:
        """Generate openai chat completion asynchronously, and parse the output.
        Example prompt is as follows:
        -----
        Given a topic, you are required to generate a story.
        You must follow the provided output format.

        Topic:
        {topic}

        Output format:
        [Story]
        ...
        [/Story]

        Now generate the output:
        """
        if functions == []:
            functions = None
        response = None
        parsed_success = True
        parse_result = None
        error_log = None
        try:
            response: ModelResponse = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                functions=functions,
                tools=tools,
                api_key=api_key,
            )
            raw_output = getattr(response.choices[0].message, "content", None)

            call_func: Optional[FunctionCall] = getattr(
                response.choices[0].message, "function_call", None
            )

            call_tools: Optional[List[ChatCompletionMessageToolCall]] = getattr(
                response.choices[0].message, "tool_calls", None
            )

            if not call_func and not call_tools:
                # function call does not appear in output
                parse_result: ParseResult = self.__parse_output_pattern__(
                    raw_output, parsing_type
                )

                # if output_keys exist & parsed_outputs does not match with output_keys -> error
                # if parse_result.error -> error
                if (
                    output_keys is not None
                    and set(parse_result.parsed_outputs.keys()) != set(output_keys)
                ) or parse_result.error:
                    parsed_success = False
                    error_log = (
                        "Output keys do not match with parsed output keys"
                        if not parse_result.error_log
                        else parse_result.error_log
                    )

            return LLMResponse(
                api_response=response,
                raw_output=raw_output,
                parsed_outputs=parse_result.parsed_outputs if parse_result else None,
                function_call=call_func if call_func else None,
                tool_calls=call_tools if call_tools else None,
                error=not parsed_success,
                error_log=error_log,
            )
        except Exception as e:
            if response is not None:
                return LLMResponse(api_response=response, error=True, error_log=str(e))
            else:
                return LLMResponse(api_response=None, error=True, error_log=str(e))

    def stream_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
        **kwargs,
    ) -> Generator[LLMStreamResponse, None, None]:
        """Parse & stream output from openai chat completion."""
        if functions == []:
            functions = None
        response = None
        try:
            if parsing_type == ParsingType.COLON.value:
                # cannot stream colon type
                yield LLMStreamResponse(
                    error=True, error_log="Cannot stream colon type"
                )
                return
            start_time = datetime.datetime.now()
            response = completion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
                tools=tools,
                api_key=api_key,
            )

            parsed_outputs = {}
            error_occurs = False
            error_log = None

            if (functions and len(functions) > 0) or (tools and len(tools) > 0):
                # if function exists, cannot parsing in stream time
                # just stream raw output and parse after stream
                streamed_outputs = {
                    "content": "",
                    "function_call": None,
                    "api_response": None,
                }
                response_with_api_res = None
                for chunk in self.__llm_stream_response_generator__(
                    messages, response, start_time, functions, tools
                ):
                    if chunk.raw_output:
                        streamed_outputs["content"] += chunk.raw_output
                    if chunk.function_call:
                        streamed_outputs["function_call"] = chunk.function_call
                    if (
                        chunk.api_response
                        and getattr(chunk.api_response.choices[0], "delta", None)
                        is None
                    ):  # only get the last api_response, not delta response
                        streamed_outputs["api_response"] = chunk.api_response
                        response_with_api_res = chunk
                    else:
                        yield chunk

                    if chunk.error and not error_occurs:
                        error_occurs = True
                        error_log = chunk.error_log

                if not streamed_outputs["function_call"]:
                    # if function call does not exist in output
                    # able to parse
                    parse_result: ParseResult = self.__parse_output_pattern__(
                        streamed_outputs["content"], parsing_type
                    )

                    error_occurs = parse_result.error or error_occurs
                    error_log = parse_result.error_log if not error_log else error_log

                    if (
                        output_keys is not None
                        and set(parse_result.parsed_outputs.keys()) != set(output_keys)
                    ) or error_occurs:
                        error_occurs = True
                        error_log = (
                            "Output keys do not match with parsed output keys"
                            if not error_log
                            else error_log
                        )
                        yield LLMStreamResponse(
                            api_response=streamed_outputs["api_response"],
                            error=True,
                            error_log=error_log,
                        )
                    else:
                        response_with_api_res.parsed_outputs = (
                            parse_result.parsed_outputs
                        )
                        yield response_with_api_res
                else:
                    yield response_with_api_res
            else:
                if parsing_type is None:
                    for chunk in self.__llm_stream_response_generator__(
                        messages, response, start_time, functions, tools
                    ):
                        yield chunk

                        if chunk.error and not error_occurs:
                            error_occurs = True
                            error_log = chunk.error_log

                elif parsing_type == ParsingType.DOUBLE_SQUARE_BRACKET.value:
                    for chunk in self.__double_type_sp_generator__(
                        messages, response, parsing_type, start_time, functions, tools
                    ):
                        yield chunk
                        if chunk.parsed_outputs:
                            parsed_outputs = update_dict(
                                parsed_outputs, chunk.parsed_outputs
                            )
                        if chunk.error and not error_occurs:
                            error_occurs = True
                            error_log = chunk.error_log
                else:
                    for chunk in self.__single_type_sp_generator__(
                        messages, response, parsing_type, start_time, functions, tools
                    ):
                        yield chunk
                        if chunk.parsed_outputs:
                            parsed_outputs = update_dict(
                                parsed_outputs, chunk.parsed_outputs
                            )
                        if chunk.error and not error_occurs:
                            error_occurs = True
                            error_log = chunk.error_log

                if (
                    output_keys is not None
                    and set(parsed_outputs.keys()) != set(output_keys)
                ) and not error_occurs:
                    error_occurs = True
                    error_log = "Output keys do not match with parsed output keys"
                    yield LLMStreamResponse(error=True, error_log=error_log)

        except Exception as e:
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def astream_and_parse(
        self,
        messages: List[Dict[str, str]],
        parsing_type: Optional[ParsingType] = None,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
        output_keys: Optional[List[str]] = None,
        model: Optional[str] = DEFAULT_MODEL,
        api_key: Optional[str] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        """Parse & stream output from openai chat completion."""
        if functions == []:
            functions = None
        response = None
        try:
            if parsing_type == ParsingType.COLON.value:
                # cannot stream colon type
                yield LLMStreamResponse(
                    error=True, error_log="Cannot stream colon type"
                )
                return
            start_time = datetime.datetime.now()
            response = await acompletion(
                model=model,
                messages=[
                    message.model_dump(exclude_none=True)
                    for message in self.__validate_openai_messages(messages)
                ],
                stream=True,
                functions=functions,
                tools=tools,
                api_key=api_key,
            )

            parsed_outputs = {}
            error_occurs = False  # error in stream time
            error_log = None
            if (functions and len(functions) > 0) or (tools and len(tools) > 0):
                # if function exists, cannot parsing in stream time
                # just stream raw output and parse after stream
                streamed_outputs = {
                    "content": "",
                    "function_call": None,
                    "api_response": None,
                }
                response_with_api_res = None
                async for chunk in self.__llm_stream_response_agenerator__(
                    messages, response, start_time, functions, tools
                ):
                    if chunk.raw_output:
                        streamed_outputs["content"] += chunk.raw_output
                    if chunk.function_call:
                        streamed_outputs["function_call"] = chunk.function_call
                    if (
                        chunk.api_response
                        and getattr(chunk.api_response.choices[0], "delta", None)
                        is None
                    ):
                        streamed_outputs["api_response"] = chunk.api_response
                        response_with_api_res = chunk
                    else:
                        yield chunk

                    if chunk.error and not error_occurs:
                        error_occurs = True
                        error_log = chunk.error_log

                if not streamed_outputs["function_call"]:
                    # if function call does not exist in output
                    # able to parse
                    parse_result: ParseResult = self.__parse_output_pattern__(
                        streamed_outputs["content"], parsing_type
                    )

                    error_occurs = parse_result.error or error_occurs
                    error_log = parse_result.error_log if not error_log else error_log
                    if (
                        output_keys is not None
                        and set(parse_result.parsed_outputs.keys()) != set(output_keys)
                    ) or error_occurs:
                        error_occurs = True
                        error_log = (
                            "Output keys do not match with parsed output keys"
                            if not error_log
                            else error_log
                        )
                        yield LLMStreamResponse(
                            api_response=streamed_outputs["api_response"],
                            error=True,
                            error_log=error_log,
                        )
                    else:
                        response_with_api_res.parsed_outputs = (
                            parse_result.parsed_outputs
                        )
                        yield response_with_api_res
                else:
                    yield response_with_api_res
            else:
                if parsing_type is None:
                    async for chunk in self.__llm_stream_response_agenerator__(
                        messages, response, start_time, functions, tools
                    ):
                        yield chunk

                        if chunk.error and not error_occurs:
                            error_occurs = True
                            error_log = chunk.error_log

                elif parsing_type == ParsingType.DOUBLE_SQUARE_BRACKET.value:
                    async for chunk in self.__double_type_sp_agenerator__(
                        messages, response, parsing_type, start_time, functions, tools
                    ):
                        yield chunk
                        if chunk.parsed_outputs:
                            parsed_outputs = update_dict(
                                parsed_outputs, chunk.parsed_outputs
                            )
                        if chunk.error and not error_occurs:
                            error_occurs = True
                else:
                    async for chunk in self.__single_type_sp_agenerator__(
                        messages, response, parsing_type, start_time, functions, tools
                    ):
                        yield chunk
                        if chunk.parsed_outputs:
                            parsed_outputs = update_dict(
                                parsed_outputs, chunk.parsed_outputs
                            )
                        if chunk.error and not error_occurs:
                            error_occurs = True

                if (
                    output_keys is not None
                    and set(parsed_outputs.keys()) != set(output_keys)
                ) and not error_occurs:
                    error_occurs = True
                    error_log = "Output keys do not match with parsed output keys"
                    yield LLMStreamResponse(error=True, error_log=error_log)

        except Exception as e:
            yield LLMStreamResponse(error=True, error_log=str(e))

    def make_model_response(
        self,
        chunk: ModelResponse,
        response_ms,
        messages: List[Dict[str, str]],
        raw_output: str,
        functions: Optional[List[Any]] = None,
        function_call: Optional[Dict[str, Any]] = None,
        tools: Optional[List[Any]] = None,
        tool_calls: Optional[List[Dict[str, Any]]] = None,
    ) -> ModelResponse:
        count_start_time = datetime.datetime.now()
        prompt_token: int = num_tokens_for_messages(
            messages=messages, model=chunk["model"]
        )
        completion_token: int = num_tokens_for_messages(
            model=chunk["model"],
            messages=[{"role": "assistant", "content": raw_output}],
        )

        if functions and len(functions) > 0:
            functions_token = num_tokens_from_functions_input(
                functions=functions, model=chunk["model"]
            )
            prompt_token += functions_token

        if tools and len(tools) > 0:
            tools_token = num_tokens_from_functions_input(
                functions=[tool["function"] for tool in tools], model=chunk["model"]
            )
            prompt_token += tools_token
        # if function_call:
        #     function_call_token = num_tokens_from_function_call_output(
        #         function_call_output=function_call, model=chunk["model"]
        #     )
        #     completion_token += function_call_token

        count_end_time = datetime.datetime.now()
        logger.debug(
            f"counting token time : {(count_end_time - count_start_time).total_seconds() * 1000} ms"
        )

        usage = Usage(
            **{
                "prompt_tokens": prompt_token,
                "completion_tokens": completion_token,
                "total_tokens": prompt_token + completion_token,
            }
        )

        last_message = Message(
            role=chunk.choices[0].delta.role
            if getattr(chunk.choices[0].delta, "role", None)
            else "assistant",
            content=raw_output if raw_output != "" else None,
            function_call=function_call if function_call else None,
            tool_calls=tool_calls if tool_calls else None,
        )
        choices = [
            Choices(finish_reason=chunk.choices[0].finish_reason, message=last_message)
        ]

        res = ModelResponse(
            id=chunk["id"],
            created=chunk["created"],
            model=chunk["model"],
            stream=True,
        )
        res.choices = choices
        res.usage = usage
        res._response_ms = response_ms

        return res

    def __llm_stream_response_generator__(
        self,
        messages: List[Dict[str, str]],
        response: Generator[ModelResponse, None, None],
        start_time: datetime.datetime,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
    ) -> Generator[LLMStreamResponse, None, None]:
        raw_output = ""
        function_call = {"name": "", "arguments": ""}
        tool_calls = []

        try:
            for chunk in response:
                yield_api_response_with_fc = False
                if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                    for key, value in (
                        chunk.choices[0].delta.function_call.model_dump().items()
                    ):
                        if value is not None:
                            function_call[key] += value

                    yield LLMStreamResponse(
                        api_response=chunk,
                        function_call=chunk.choices[0].delta.function_call,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "tool_calls", None) is not None:
                    # tool_calls: list
                    tool_calls_delta: List[Any] = chunk.choices[0].delta.tool_calls
                    index = tool_calls_delta[0].index
                    if index == len(tool_calls):
                        tool_calls.append(
                            {
                                "id": tool_calls_delta[0].id,
                                "function": {},
                                "type": "function",
                            }
                        )
                    tool_delta: ChoiceDeltaToolCallFunction = tool_calls_delta[
                        0
                    ].function
                    tool_calls[index]["function"] = update_dict(
                        tool_calls[index]["function"], tool_delta.model_dump()
                    )

                    yield LLMStreamResponse(
                        api_response=chunk,
                        tool_calls=chunk.choices[0].delta.tool_calls,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "content", None) is not None:
                    raw_output += chunk.choices[0].delta.content
                    yield LLMStreamResponse(
                        api_response=chunk if not yield_api_response_with_fc else None,
                        raw_output=chunk.choices[0].delta.content,
                    )

                if chunk.choices[0].finish_reason != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            functions=functions,
                            function_call=function_call
                            if chunk.choices[0].finish_reason == "function_call"
                            else None,
                            tools=tools,
                            tool_calls=tool_calls
                            if chunk.choices[0].finish_reason == "tool_calls"
                            else None,
                        )
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    def __single_type_sp_generator__(
        self,
        messages: List[Dict[str, str]],
        response: Generator[ModelResponse, None, None],
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
    ) -> Generator[LLMStreamResponse, None, None]:
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            tool_calls = []

            for chunk in response:
                yield_api_response_with_fc = False
                if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                    for key, value in (
                        chunk.choices[0].delta.function_call.model_dump().items()
                    ):
                        if value is not None:
                            function_call[key] += value

                    yield LLMStreamResponse(
                        api_response=chunk,
                        function_call=chunk.choices[0].delta.function_call,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "tool_calls", None) is not None:
                    # tool_calls: list
                    tool_calls_delta: List[Any] = chunk.choices[0].delta.tool_calls
                    index = tool_calls_delta[0].index
                    if index == len(tool_calls):
                        tool_calls.append(
                            {
                                "id": tool_calls_delta[0].id,
                                "function": {},
                                "type": "function",
                            }
                        )
                    tool_delta: ChoiceDeltaToolCallFunction = tool_calls_delta[
                        0
                    ].function
                    tool_calls[index]["function"] = update_dict(
                        tool_calls[index]["function"], tool_delta.model_dump()
                    )

                    yield LLMStreamResponse(
                        api_response=chunk,
                        tool_calls=chunk.choices[0].delta.tool_calls,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "content", None) is not None:
                    stream_value: str = chunk.choices[0].delta.content
                    raw_output += stream_value
                    yield LLMStreamResponse(
                        api_response=chunk if not yield_api_response_with_fc else None,
                        raw_output=stream_value,
                    )

                    buffer += stream_value
                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            if len(keys) == 0:
                                break  # no key
                            active_key, active_type = keys[
                                0
                            ]  # Updated to unpack both key and type
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]
                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[
                                                0
                                            ].replace(end_tag, "")
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if pattern ends  = ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        active_type == "List"
                                        or active_type == "Dict"
                                        and end_token.find("]") != -1
                                    ):
                                        try:
                                            buffer_dict = json.loads(buffer)
                                            stream_pause = False
                                            continue
                                        except Exception as exception:
                                            logger.error(exception)
                                            yield LLMStreamResponse(
                                                error=True,
                                                error_log="Parsing error : Invalid end tag detected",
                                                parsed_outputs={
                                                    active_key: buffer.split(
                                                        start_token
                                                    )[0]
                                                },
                                            )
                                            stream_pause = False
                                            buffer = ""
                                    yield LLMStreamResponse(
                                        error=True,
                                        error_log="Parsing error : Invalid end tag detected",
                                        parsed_outputs={active_key: buffer},
                                    )
                                    stream_pause = False
                                    buffer = ""
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if chunk.choices[0].finish_reason != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            functions=functions,
                            function_call=function_call
                            if chunk.choices[0].finish_reason == "function_call"
                            else None,
                            tools=tools,
                            tool_calls=tool_calls
                            if chunk.choices[0].finish_reason == "tool_calls"
                            else None,
                        )
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    def __double_type_sp_generator__(
        self,
        messages: List[Dict[str, str]],
        response: Generator[ModelResponse, None, None],
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
    ) -> Generator[LLMStreamResponse, None, None]:
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            tool_calls = []

            for chunk in response:
                yield_api_response_with_fc = False
                if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                    for key, value in (
                        chunk.choices[0].delta.function_call.model_dump().items()
                    ):
                        if value is not None:
                            function_call[key] += value

                    yield LLMStreamResponse(
                        api_response=chunk,
                        function_call=chunk.choices[0].delta.function_call,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "tool_calls", None) is not None:
                    # tool_calls: list
                    tool_calls_delta: List[Any] = chunk.choices[0].delta.tool_calls
                    index = tool_calls_delta[0].index
                    if index == len(tool_calls):
                        tool_calls.append(
                            {
                                "id": tool_calls_delta[0].id,
                                "function": {},
                                "type": "function",
                            }
                        )
                    tool_delta: ChoiceDeltaToolCallFunction = tool_calls_delta[
                        0
                    ].function
                    tool_calls[index]["function"] = update_dict(
                        tool_calls[index]["function"], tool_delta.model_dump()
                    )

                    yield LLMStreamResponse(
                        api_response=chunk,
                        tool_calls=chunk.choices[0].delta.tool_calls,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "content", None) is not None:
                    stream_value: str = chunk.choices[0].delta.content
                    raw_output += stream_value
                    yield LLMStreamResponse(
                        api_response=chunk if not yield_api_response_with_fc else None,
                        raw_output=stream_value,
                    )

                    buffer += stream_value

                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            if len(keys) == 0:
                                break  # no key
                            active_key, active_type = keys[0]
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]

                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[0]
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        buffer.find(end_token + end_token) != -1
                                    ):  # if ]] in buffer -> error
                                        yield LLMStreamResponse(
                                            error=True,
                                            error_log="Parsing error : Invalid end tag detected",
                                            parsed_outputs={
                                                active_key: buffer.split(start_token)[0]
                                            },
                                        )
                                        buffer = buffer.split(end_token + end_token)[-1]
                                        stream_pause = False
                                        break
                                    else:
                                        if (
                                            buffer.find(start_token + start_token) != -1
                                        ):  # if [[ in buffer -> pause
                                            break
                                        else:
                                            # if [ in buffer (== [blah]) -> stream
                                            yield LLMStreamResponse(
                                                parsed_outputs={active_key: buffer}
                                            )
                                            buffer = ""
                                            stream_pause = False
                                            break
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if chunk.choices[0].finish_reason != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            functions=functions,
                            function_call=function_call
                            if chunk.choices[0].finish_reason == "function_call"
                            else None,
                            tools=tools,
                            tool_calls=tool_calls
                            if chunk.choices[0].finish_reason == "tool_calls"
                            else None,
                        )
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def __llm_stream_response_agenerator__(
        self,
        messages: List[Dict[str, str]],
        response: AsyncGenerator[ModelResponse, None],
        start_time: datetime.datetime,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        raw_output = ""
        function_call = {"name": "", "arguments": ""}
        tool_calls = []
        try:
            async for chunk in response:
                yield_api_response_with_fc = False
                if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                    for key, value in (
                        chunk.choices[0].delta.function_call.model_dump().items()
                    ):
                        if value is not None:
                            function_call[key] += value

                    yield LLMStreamResponse(
                        api_response=chunk,
                        function_call=chunk.choices[0].delta.function_call,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "tool_calls", None) is not None:
                    # tool_calls: list
                    tool_calls_delta: List[Any] = chunk.choices[0].delta.tool_calls
                    index = tool_calls_delta[0].index
                    if index == len(tool_calls):
                        tool_calls.append(
                            {
                                "id": tool_calls_delta[0].id,
                                "function": {},
                                "type": "function",
                            }
                        )
                    tool_delta: ChoiceDeltaToolCallFunction = tool_calls_delta[
                        0
                    ].function
                    tool_calls[index]["function"] = update_dict(
                        tool_calls[index]["function"], tool_delta.model_dump()
                    )

                    yield LLMStreamResponse(
                        api_response=chunk,
                        tool_calls=chunk.choices[0].delta.tool_calls,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "content", None) is not None:
                    stream_value: str = chunk.choices[0].delta.content
                    raw_output += stream_value
                    yield LLMStreamResponse(
                        api_response=chunk if not yield_api_response_with_fc else None,
                        raw_output=stream_value,
                    )

                if chunk.choices[0].finish_reason != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            functions=functions,
                            function_call=function_call
                            if chunk.choices[0].finish_reason == "function_call"
                            else None,
                            tools=tools,
                            tool_calls=tool_calls
                            if chunk.choices[0].finish_reason == "tool_calls"
                            else None,
                        )
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def __single_type_sp_agenerator__(
        self,
        messages: List[Dict[str, str]],
        response: AsyncGenerator[ModelResponse, None],
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            tool_calls = []

            async for chunk in response:
                yield_api_response_with_fc = False
                if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                    for key, value in (
                        chunk.choices[0].delta.function_call.model_dump().items()
                    ):
                        if value is not None:
                            function_call[key] += value

                    yield LLMStreamResponse(
                        api_response=chunk,
                        function_call=chunk.choices[0].delta.function_call,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "tool_calls", None) is not None:
                    # tool_calls: list
                    tool_calls_delta: List[Any] = chunk.choices[0].delta.tool_calls
                    index = tool_calls_delta[0].index
                    if index == len(tool_calls):
                        tool_calls.append(
                            {
                                "id": tool_calls_delta[0].id,
                                "function": {},
                                "type": "function",
                            }
                        )
                    tool_delta: ChoiceDeltaToolCallFunction = tool_calls_delta[
                        0
                    ].function
                    tool_calls[index]["function"] = update_dict(
                        tool_calls[index]["function"], tool_delta.model_dump()
                    )

                    yield LLMStreamResponse(
                        api_response=chunk,
                        tool_calls=chunk.choices[0].delta.tool_calls,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "content", None) is not None:
                    stream_value: str = chunk.choices[0].delta.content
                    raw_output += stream_value
                    yield LLMStreamResponse(
                        api_response=chunk if not yield_api_response_with_fc else None,
                        raw_output=stream_value,
                    )

                    buffer += stream_value

                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            if len(keys) == 0:
                                break  # no key

                            active_key, active_type = keys[
                                0
                            ]  # Updated to unpack both key and type
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]

                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[
                                                0
                                            ].replace(end_tag, "")
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if pattern ends  = ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        active_type == "List"
                                        or active_type == "Dict"
                                        and end_token.find("]") != -1
                                    ):
                                        try:
                                            buffer_dict = json.loads(buffer)
                                            stream_pause = False
                                            continue
                                        except Exception as exception:
                                            logger.error(exception)
                                            yield LLMStreamResponse(
                                                error=True,
                                                error_log="Parsing error : Invalid end tag detected",
                                                parsed_outputs={
                                                    active_key: buffer.split(
                                                        start_token
                                                    )[0]
                                                },
                                            )
                                            stream_pause = False
                                            buffer = ""
                                    yield LLMStreamResponse(
                                        error=True,
                                        error_log="Parsing error : Invalid end tag detected",
                                        parsed_outputs={active_key: buffer},
                                    )
                                    stream_pause = False
                                    buffer = ""
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if chunk.choices[0].finish_reason != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            functions=functions,
                            function_call=function_call
                            if chunk.choices[0].finish_reason == "function_call"
                            else None,
                            tools=tools,
                            tool_calls=tool_calls
                            if chunk.choices[0].finish_reason == "tool_calls"
                            else None,
                        )
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))

    async def __double_type_sp_agenerator__(
        self,
        messages: List[Dict[str, str]],
        response: AsyncGenerator[ModelResponse, None],
        parsing_type: ParsingType,
        start_time: datetime.datetime,
        functions: Optional[List[Any]] = None,
        tools: Optional[List[Any]] = None,
    ) -> AsyncGenerator[LLMStreamResponse, None]:
        try:
            parsing_pattern = get_pattern_by_type(parsing_type)
            start_tag = parsing_pattern["start"]
            start_fstring = parsing_pattern["start_fstring"]
            end_fstring = parsing_pattern["end_fstring"]
            start_token = parsing_pattern["start_token"]
            end_token = parsing_pattern["end_token"]

            buffer = ""
            raw_output = ""
            active_key = None
            stream_pause = False
            end_tag = None
            function_call = {"name": "", "arguments": ""}
            tool_calls = []

            async for chunk in response:
                yield_api_response_with_fc = False
                if getattr(chunk.choices[0].delta, "function_call", None) is not None:
                    for key, value in (
                        chunk.choices[0].delta.function_call.model_dump().items()
                    ):
                        if value is not None:
                            function_call[key] += value

                    yield LLMStreamResponse(
                        api_response=chunk,
                        function_call=chunk.choices[0].delta.function_call,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "tool_calls", None) is not None:
                    # tool_calls: list
                    tool_calls_delta: List[Any] = chunk.choices[0].delta.tool_calls
                    index = tool_calls_delta[0].index
                    if index == len(tool_calls):
                        tool_calls.append(
                            {
                                "id": tool_calls_delta[0].id,
                                "function": {},
                                "type": "function",
                            }
                        )
                    tool_delta: ChoiceDeltaToolCallFunction = tool_calls_delta[
                        0
                    ].function
                    tool_calls[index]["function"] = update_dict(
                        tool_calls[index]["function"], tool_delta.model_dump()
                    )

                    yield LLMStreamResponse(
                        api_response=chunk,
                        tool_calls=chunk.choices[0].delta.tool_calls,
                    )
                    yield_api_response_with_fc = True

                if getattr(chunk.choices[0].delta, "content", None) is not None:
                    stream_value: str = chunk.choices[0].delta.content
                    raw_output += stream_value
                    yield LLMStreamResponse(
                        api_response=chunk if not yield_api_response_with_fc else None,
                        raw_output=stream_value,
                    )

                    buffer += stream_value

                    while True:
                        if active_key is None:
                            keys = re.findall(start_tag, buffer, flags=re.DOTALL)
                            # if len(keys) > 1:
                            #     yield LLMStreamResponse(
                            #         error=True,
                            #         error_log="Parsing error : Nested key detected",
                            #     )
                            #     break
                            if len(keys) == 0:
                                break  # no key
                            active_key, active_type = keys[0]
                            end_tag = end_fstring.format(key=active_key)
                            # delete start tag from buffer
                            start_pattern = start_fstring.format(
                                key=active_key, type=active_type
                            )
                            buffer = buffer.split(start_pattern)[-1]

                        else:
                            if (
                                stream_value.find(start_token) != -1
                            ):  # start token appers in chunk -> pause
                                stream_pause = True
                                break
                            elif stream_pause:
                                if (
                                    buffer.find(end_tag) != -1
                                ):  # if end tag appears in buffer
                                    yield LLMStreamResponse(
                                        parsed_outputs={
                                            active_key: buffer.split(end_tag)[0]
                                        }
                                    )
                                    buffer = buffer.split(end_tag)[-1]
                                    active_key = None
                                    stream_pause = False
                                    # break
                                elif (
                                    stream_value.find(end_token) != -1
                                ):  # if ("[blah]" != end_pattern) appeared in buffer
                                    if (
                                        buffer.find(end_token + end_token) != -1
                                    ):  # if ]] in buffer -> error
                                        yield LLMStreamResponse(
                                            error=True,
                                            error_log="Parsing error : Invalid end tag detected",
                                            parsed_outputs={
                                                active_key: buffer.split(start_token)[0]
                                            },
                                        )
                                        buffer = buffer.split(end_token + end_token)[-1]
                                        stream_pause = False
                                        break
                                    else:
                                        if (
                                            buffer.find(start_token + start_token) != -1
                                        ):  # if [[ in buffer -> pause
                                            break
                                        else:
                                            # if [ in buffer (== [blah]) -> stream
                                            yield LLMStreamResponse(
                                                parsed_outputs={active_key: buffer}
                                            )
                                            buffer = ""
                                            stream_pause = False
                                            break
                                break
                            else:
                                # no start token, no stream_pause (not inside of tag)
                                if buffer:
                                    yield LLMStreamResponse(
                                        parsed_outputs={active_key: buffer}
                                    )
                                    buffer = ""
                                break

                if chunk.choices[0].finish_reason != None:
                    end_time = datetime.datetime.now()
                    response_ms = (end_time - start_time).total_seconds() * 1000
                    yield LLMStreamResponse(
                        api_response=self.make_model_response(
                            chunk,
                            response_ms,
                            messages,
                            raw_output,
                            functions=functions,
                            function_call=function_call
                            if chunk.choices[0].finish_reason == "function_call"
                            else None,
                            tools=tools,
                            tool_calls=tool_calls
                            if chunk.choices[0].finish_reason == "tool_calls"
                            else None,
                        )
                    )
        except Exception as e:
            logger.error(e)
            yield LLMStreamResponse(error=True, error_log=str(e))



import asyncio
import json
import datetime
import re

from uuid import UUID, uuid4
from typing import Dict, Any, Optional, AsyncGenerator, List
from dotenv import load_dotenv
from collections import defaultdict
from asyncio import Queue

from websockets.client import connect, WebSocketClientProtocol
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from readerwriterlock import rwlock
from playhouse.shortcuts import model_to_dict

import promptmodel.utils.logger as logger
from promptmodel import DevApp
from promptmodel.llms.llm_dev import LLMDev
from promptmodel.database.models import (
    DeployedFunctionModel,
    DeployedFunctionModelVersion,
    DeployedPrompt,
)
from promptmodel.types.enums import ServerTask, LocalTask, LocalTaskErrorType
from promptmodel.utils.config_utils import upsert_config, read_config
from promptmodel.utils.output_utils import update_dict
from promptmodel.types.response import LLMStreamResponse
from promptmodel.constants import ENDPOINT_URL

load_dotenv()
GATEWAY_URL = f"wss://{ENDPOINT_URL.split('://')[1]}/open_websocket"


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            aware_datetime = obj.replace(tzinfo=datetime.timezone.utc)
            return aware_datetime.isoformat()  # This will include timezone information
        return super().default(obj)


class DevWebsocketClient:
    def __init__(self, _devapp: DevApp):
        self._devapp: DevApp = _devapp
        self.rwlock = rwlock.RWLockFair()
        self.pending_requests: Dict[str, asyncio.Event] = {}
        self.responses: Dict[str, Queue] = defaultdict(Queue)

    async def _get_function_models(self, function_model_name: str):
        """Get function_model from registry"""
        with self.rwlock.gen_rlock():
            function_model = next(
                (
                    function_model
                    for function_model in self._devapp.function_models
                    if function_model.name == function_model_name
                ),
                None,
            )
        return function_model

    def update_devapp_instance(self, new_devapp):
        with self.rwlock.gen_wlock():
            self._devapp = new_devapp

    async def __handle_message(
        self, message: Dict[str, Any], ws: WebSocketClientProtocol
    ):
        # logger.info(f"Received message: {message}")
        response: Dict[Any, str] = {}
        # If the message has a correlation_id, add it to the response
        # correlation_id is the unique ID of the function from backend to local
        if message.get("correlation_id"):
            response["correlation_id"] = message["correlation_id"]

        # If the message has a runner_id, add it to the response
        if message.get("runner_id"):
            response["runner_id"] = message["runner_id"]

        data = None
        try:
            if message["type"] == LocalTask.RUN_PROMPT_MODEL:
                messages: List[Dict] = message["messages_for_run"]

                # # Check function_model in Local Usage
                # function_model_names = self._devapp._get_function_model_name_list()
                # if function_model_name not in function_model_names:
                #     logger.error(f"There is no function_model {function_model_name}.")
                #     return

                # Start FunctionModel Running
                output = {"raw_output": "", "parsed_outputs": {}}
                try:
                    logger.info("Started FunctionModel")
                    # create function_model_dev_instance
                    function_model_dev = LLMDev()
                    # find function_model_uuid from local db

                    data = {
                        "type": ServerTask.UPDATE_RESULT_RUN.value,
                        "status": "running",
                    }
                    data.update(response)
                    # logger.debug(f"Sent response: {data}")
                    await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    model = message["model"]
                    parsing_type = message["parsing_type"]

                    messages_for_run = messages

                    parsing_success = True
                    error_log = None
                    function_call = None
                    function_schemas: Optional[List[Dict]] = (
                        message["function_schemas"]
                        if "function_schemas" in message
                        else None
                    )  # ehese schemata have mock_response which should not be sent to LLM
                    function_mock_responses = {}

                    if function_schemas:
                        for function_schema in function_schemas:
                            function_mock_responses[
                                function_schema["name"]
                            ] = function_schema["mock_response"]

                        for schema in function_schemas:
                            del schema["mock_response"]

                    res: AsyncGenerator[
                        LLMStreamResponse, None
                    ] = function_model_dev.dev_run(
                        messages=messages_for_run,
                        parsing_type=parsing_type,
                        functions=function_schemas,
                        model=model,
                    )

                    async for item in res:
                        # send item to backend
                        # save item & parse
                        # if type(item) == str: raw output, if type(item) == dict: parsed output
                        if item.raw_output is not None:
                            output["raw_output"] += item.raw_output
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "raw_output": item.raw_output,
                            }
                        if item.parsed_outputs:
                            output["parsed_outputs"] = update_dict(
                                output["parsed_outputs"], item.parsed_outputs
                            )

                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "parsed_outputs": item.parsed_outputs,
                            }
                        if item.function_call is not None:
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "function_call": item.function_call.model_dump(),
                            }
                            function_call = item.function_call.model_dump()

                        if item.error and parsing_success is True:
                            parsing_success = not item.error
                            error_log = item.error_log

                        data.update(response)
                        # logger.debug(f"Sent response: {data}")
                        await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    # IF function_call in response -> call function
                    if function_call:
                        if (
                            function_call["name"]
                            in self._devapp._get_function_name_list()
                        ):
                            # call function
                            try:
                                function_call_args: Dict[str, Any] = json.loads(
                                    function_call["arguments"]
                                )
                                function_response = (
                                    self._devapp._call_register_function(
                                        function_call["name"], function_call_args
                                    )
                                )

                                # Send function call response for check LLM response validity
                                data = {
                                    "type": ServerTask.UPDATE_RESULT_RUN.value,
                                    "status": "running",
                                    "function_response": {
                                        "name": function_call["name"],
                                        "response": function_response,
                                    },
                                }
                                data.update(response)
                                # logger.debug(f"Sent response: {data}")
                                await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                            except Exception as error:
                                logger.error(f"{error}")

                                data = {
                                    "type": ServerTask.UPDATE_RESULT_RUN.value,
                                    "status": "failed",
                                    "error_type": LocalTaskErrorType.FUNCTION_CALL_FAILED_ERROR.value,
                                    "log": f"Function call Failed, {error}",
                                }

                                response.update(data)
                                await ws.send(
                                    json.dumps(response, cls=CustomJSONEncoder)
                                )
                                return

                        else:
                            # return mock response
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "function_response": {
                                    "name": function_call["name"],
                                    "response": "FAKE RESPONSE : "
                                    + str(
                                        function_mock_responses[function_call["name"]]
                                    ),
                                },
                            }
                            data.update(response)
                            # logger.debug(f"Sent response: {data}")
                            await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    if (
                        message["output_keys"] is not None
                        and message["parsing_type"] is not None
                        and set(output["parsed_outputs"].keys())
                        != set(
                            message["output_keys"]
                        )  # parsed output keys != output keys
                    ) or (
                        parsing_success is False
                    ):  # error occurs in streaming time
                        error_log = error_log if error_log else "Key matching failed."
                        data = {
                            "type": ServerTask.UPDATE_RESULT_RUN.value,
                            "status": "failed",
                            "error_type": LocalTaskErrorType.PARSING_FAILED_ERROR.value,
                            "log": f"parsing failed, {error_log}",
                        }
                        response.update(data)
                        await ws.send(json.dumps(response, cls=CustomJSONEncoder))
                        return

                    data = {
                        "type": ServerTask.UPDATE_RESULT_RUN.value,
                        "status": "completed",
                    }

                except Exception as error:
                    logger.error(f"Error running service: {error}")
                    data = {
                        "type": ServerTask.UPDATE_RESULT_RUN.value,
                        "status": "failed",
                        "error_type": LocalTaskErrorType.SERVICE_ERROR.value,
                        "log": str(error),
                    }
                    response.update(data)
                    await ws.send(json.dumps(response, cls=CustomJSONEncoder))
                    return

            elif message["type"] == LocalTask.RUN_CHAT_MODEL:
                old_messages = message["old_messages"]
                new_messages = message["new_messages"]
                # move tool_calls in old_messages into new_messages
                for old_message in old_messages:
                    if "tool_calls" in old_message:
                        if type(old_message["tool_calls"]) == List:
                            old_message["function_call"] = old_message["tool_calls"][0]
                        elif type(old_message["tool_calls"]) == Dict:
                            old_message["function_call"] = old_message["tool_calls"]
                        del old_message["tool_calls"]

                # Start ChatModel Running
                try:
                    logger.info("Started ChatModel")
                    chat_model_dev = LLMDev()

                    messages_for_run = old_messages + new_messages

                    error_log = None
                    function_call = None

                    function_schemas: Optional[List[Dict]] = (
                        message["function_schemas"]
                        if "function_schemas" in message
                        else None
                    )  # this has a mock_response which should not be sent to LLM
                    function_mock_responses = {}

                    if function_schemas:
                        for function_schema in function_schemas:
                            function_mock_responses[
                                function_schema["name"]
                            ] = function_schema["mock_response"]

                        for schema in function_schemas:
                            del schema["mock_response"]

                    res: AsyncGenerator[
                        LLMStreamResponse, None
                    ] = chat_model_dev.dev_chat(
                        messages=messages_for_run,
                        functions=function_schemas,
                        model=message["model"],
                    )

                    raw_output = ""
                    async for chunk in res:
                        logger.debug(f"Chunk: {chunk}")
                        if chunk.raw_output is not None:
                            raw_output += chunk.raw_output
                            data = {
                                "type": ServerTask.UPDATE_RESULT_CHAT_RUN.value,
                                "status": "running",
                                "raw_output": chunk.raw_output,
                            }
                        if chunk.function_call is not None:
                            data = {
                                "type": ServerTask.UPDATE_RESULT_CHAT_RUN.value,
                                "status": "running",
                                "function_call": chunk.function_call.model_dump(),
                            }
                            if function_call is None:
                                function_call = {}
                            function_call = update_dict(
                                function_call, chunk.function_call.model_dump()
                            )

                        if chunk.error:
                            error_log = chunk.error_log

                        data.update(response)
                        # logger.debug(f"Sent response: {data}")
                        await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                    # IF function_call in response -> call function -> call LLM once more
                    logger.debug(f"Function call: {function_call}")

                    if function_call is not None:
                        if (
                            function_call["name"]
                            in self._devapp._get_function_name_list()
                        ):
                            # call function
                            try:
                                function_call_args: Dict[str, Any] = json.loads(
                                    function_call["arguments"]
                                )
                                function_response = (
                                    self._devapp._call_register_function(
                                        function_call["name"], function_call_args
                                    )
                                )

                                # Send function call response for check LLM response validity
                                data = {
                                    "type": ServerTask.UPDATE_RESULT_CHAT_RUN.value,
                                    "status": "running",
                                    "function_response": {
                                        "name": function_call["name"],
                                        "response": function_response,
                                    },
                                }
                                data.update(response)
                                await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                                logger.debug(f"Sent response: {data}")
                            except Exception as error:
                                logger.error(f"{error}")

                                data = {
                                    "type": ServerTask.UPDATE_RESULT_CHAT_RUN.value,
                                    "status": "failed",
                                    "error_type": LocalTaskErrorType.FUNCTION_CALL_FAILED_ERROR.value,
                                    "log": f"Function call Failed, {error}",
                                }

                                response.update(data)
                                await ws.send(
                                    json.dumps(response, cls=CustomJSONEncoder)
                                )
                                return
                        else:
                            # return mock response
                            data = {
                                "type": ServerTask.UPDATE_RESULT_RUN.value,
                                "status": "running",
                                "function_response": {
                                    "name": function_call["name"],
                                    "response": "FAKE RESPONSE : "
                                    + function_mock_responses[function_call["name"]],
                                },
                            }
                            data.update(response)
                            # logger.debug(f"Sent response: {data}")
                            await ws.send(json.dumps(data, cls=CustomJSONEncoder))
                            function_response = function_mock_responses[
                                function_call["name"]
                            ]

                        # call LLM once more
                        messages_for_run += [
                            {
                                "role": "assistant",
                                "content": "",
                                "function_call": function_call,
                            },
                            {
                                "role": "function",
                                "name": function_call["name"],
                                "content": str(function_response),
                            },
                        ]

                        res_after_function_call: AsyncGenerator[
                            LLMStreamResponse, None
                        ] = chat_model_dev.dev_chat(
                            messages=messages_for_run,
                            model=message["model"],
                        )

                        raw_output = ""
                        async for item in res_after_function_call:
                            if item.raw_output is not None:
                                raw_output += item.raw_output
                                data = {
                                    "type": ServerTask.UPDATE_RESULT_CHAT_RUN.value,
                                    "status": "running",
                                    "raw_output": item.raw_output,
                                }

                            if item.error:
                                error_log = item.error_log

                            data.update(response)
                            # logger.debug(f"Sent response: {data}")
                            await ws.send(json.dumps(data, cls=CustomJSONEncoder))

                    data = {
                        "type": ServerTask.UPDATE_RESULT_CHAT_RUN.value,
                        "status": "completed",
                    }

                except Exception as error:
                    logger.error(f"Error running service: {error}")
                    data = {
                        "type": ServerTask.UPDATE_RESULT_CHAT_RUN.value,
                        "status": "failed",
                        "error_type": LocalTaskErrorType.SERVICE_ERROR.value,
                        "log": str(error),
                    }
                    response.update(data)
                    await ws.send(json.dumps(response, cls=CustomJSONEncoder))
                    return

            if data:
                response.update(data)
                await ws.send(json.dumps(response, cls=CustomJSONEncoder))
            logger.info(f"Sent response: {response}")
        except Exception as error:
            logger.error(f"Error handling message: {error}")
            await ws.send(str(error))

    async def connect_to_gateway(
        self,
        project_uuid: str,
        connection_name: str,
        cli_access_header: dict,
        retries=12 * 24,
        retry_delay=5 * 60,
    ):
        """Open Websocket to Backend with project_uuid, connection_name, cli_access_token"""
        headers = cli_access_header
        headers.update(
            {"project_uuid": project_uuid, "connection_name": connection_name}
        )
        for _ in range(retries):
            try:
                async with connect(
                    GATEWAY_URL,
                    extra_headers=headers,
                    # ping_interval=10,
                    # ping_timeout=1,
                    # timeout=3600 * 24,  # Timeout is set to 24 hours
                ) as ws:
                    logger.success("Connected to gateway. Your DevApp is now online! 🎉")
                    self.ws = ws
                    while True:
                        message = await ws.recv()
                        data = json.loads(message)
                        correlation_id = data.get("correlation_id")

                        if correlation_id and correlation_id in self.pending_requests:
                            await self.responses[correlation_id].put(data)
                            if not self.pending_requests[correlation_id].is_set():
                                self.pending_requests[
                                    correlation_id
                                ].set()  # Signal the event that the response has arrived
                        else:
                            await self.__handle_message(data, ws)
            except (ConnectionClosedError, ConnectionClosedOK):
                # If the connection was closed gracefully, handle it accordingly
                logger.warning("Connection to the gateway was closed.")
            except TimeoutError:
                logger.error(
                    f"Timeout error while connecting to the gateway. Retrying in {retry_delay} seconds..."
                )
                await asyncio.sleep(retry_delay)
            except Exception as error:
                logger.error(f"Error receiving message: {error}")

    async def request(self, type: ServerTask, message: Dict = {}):
        """
        Send a message to the connected server and wait for a response.

        Returns a python object.
        """
        ws = self.ws
        if ws:
            correlation_id = str(uuid4())  # Generate unique correlation ID
            message["correlation_id"] = correlation_id

            try:
                message["type"] = type.value
                await ws.send(json.dumps(message))
                logger.success(
                    f"""Sent request to local.
    - Message: {message}"""
                )
                event = asyncio.Event()
                self.pending_requests[correlation_id] = event

                await asyncio.wait_for(event.wait(), timeout=120)  # 2 minutes timeout
                response = await self.responses[correlation_id].get()
                logger.debug(response)
                return response
            except Exception as error:
                logger.error(
                    f"""Error for request to local: {error}
    - Message: {message}"""
                )
            finally:
                self.pending_requests.pop(correlation_id, None)
                self.responses.pop(correlation_id, None)
        else:
            raise ValueError(f"No active connection found")



import os
import sys
import importlib
import asyncio
from typing import Any, Dict, List
from threading import Timer
from rich import print
from watchdog.events import FileSystemEventHandler
from playhouse.shortcuts import model_to_dict

from promptmodel.apis.base import APIClient
from promptmodel.utils.config_utils import read_config, upsert_config
from promptmodel.utils.async_utils import run_async_in_sync_threadsafe
from promptmodel.utils import logger
from promptmodel import DevApp
from promptmodel.database.models import (
    DeployedFunctionModel,
    DeployedFunctionModelVersion,
    DeployedPrompt,
)

from promptmodel.websocket.websocket_client import DevWebsocketClient
from promptmodel.types.enums import ServerTask


class CodeReloadHandler(FileSystemEventHandler):
    def __init__(
        self,
        _devapp_filename: str,
        _instance_name: str,
        dev_websocket_client: DevWebsocketClient,
        main_loop: asyncio.AbstractEventLoop,
    ):
        self._devapp_filename: str = _devapp_filename
        self.devapp_instance_name: str = _instance_name
        self.dev_websocket_client: DevWebsocketClient = (
            dev_websocket_client  # save dev_websocket_client instance
        )
        self.timer = None
        self.main_loop = main_loop

    def on_modified(self, event):
        """Called when a file or directory is modified."""
        if event.src_path.endswith(".py"):
            upsert_config({"reloading": True}, "connection")
            if self.timer:
                self.timer.cancel()
            # reload modified file & main file
            self.timer = Timer(0.5, self.reload_code, args=(event.src_path,))
            self.timer.start()

    def reload_code(self, modified_file_path: str):
        print(
            f"[violet]promptmodel:dev:[/violet]  Reloading {self._devapp_filename} module due to changes..."
        )
        relative_modified_path = os.path.relpath(modified_file_path, os.getcwd())
        # Reload the devapp module
        module_name = relative_modified_path.replace("./", "").replace("/", ".")[
            :-3
        ]  # assuming the file is in the PYTHONPATH

        if module_name in sys.modules:
            module = sys.modules[module_name]
            importlib.reload(module)

        reloaded_module = importlib.reload(sys.modules[self._devapp_filename])
        print(
            f"[violet]promptmodel:dev:[/violet]  {self._devapp_filename} module reloaded successfully."
        )

        new_devapp_instance: DevApp = getattr(
            reloaded_module, self.devapp_instance_name
        )

        config = read_config()
        org = config["connection"]["org"]
        project = config["connection"]["project"]

        # save samples, FunctionSchema, FunctionModel, ChatModel to cloud server by websocket ServerTask request
        new_function_model_name_list = (
            new_devapp_instance._get_function_model_name_list()
        )
        new_chat_model_name_list = new_devapp_instance._get_chat_model_name_list()
        new_samples = new_devapp_instance.samples
        new_function_schemas = new_devapp_instance._get_function_schema_list()

        res = run_async_in_sync_threadsafe(
            self.dev_websocket_client.request(
                ServerTask.SYNC_CODE,
                message={
                    "new_function_model": new_function_model_name_list,
                    "new_chat_model": new_chat_model_name_list,
                    "new_samples": new_samples,
                    "new_schemas": new_function_schemas,
                },
            ),
            main_loop=self.main_loop,
        )

        # update_samples(new_devapp_instance.samples)
        upsert_config({"reloading": False}, "connection")
        self.dev_websocket_client.update_devapp_instance(new_devapp_instance)



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from promptmodel.llms.llm_proxy import LLMProxy


async def echo_coroutine(*args, **kwargs):
    # print(args, kwargs)
    return args, kwargs


@pytest.fixture
def mock_fetch_prompts():
    mock_fetch_prompts = AsyncMock()
    mock_prompts = [
        {"role": "system", "content": "You are a helpful assistant.", "step": 1},
        {"role": "user", "content": "Hello!", "step": 2},
    ]
    mock_version_details = {
        "model": "gpt-3.5-turbo",
        "uuid": "testuuid",
        "version": 1,
        "parsing_type": None,
        "output_keys": None,
    }
    mock_fetch_prompts.return_value = (mock_prompts, mock_version_details)

    return mock_fetch_prompts


@pytest.fixture
def mock_async_log_to_cloud():
    mock_async_log_to_cloud = AsyncMock()

    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_async_log_to_cloud.return_value = mock_response

    return mock_async_log_to_cloud



import pytest
from unittest.mock import AsyncMock, patch, MagicMock

from typing import Generator, AsyncGenerator, Dict, List, Any, Optional
from litellm import ModelResponse

from promptmodel.llms.llm import LLM, ParseResult
from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.types.response import LLMResponse, LLMStreamResponse
from promptmodel.types.enums import ParsingType

proxy = LLMProxy(name="test")


def test_chat_run(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )

    res: LLMResponse = proxy.chat_run(session_uuid="testuuid")

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None
    if isinstance(res.api_response.usage, dict):
        assert res.api_response.usage["prompt_tokens"] > 15
        print(res.api_response.usage["prompt_tokens"])
    else:
        assert res.api_response.usage.prompt_tokens > 15
        print(res.api_response.usage.prompt_tokens)


@pytest.mark.asyncio
async def test_chat_arun(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )

    res: LLMResponse = await proxy.chat_arun(session_uuid="testuuid")

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None


def test_chat_stream(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )

    res: Generator[LLMStreamResponse, None, None] = proxy.chat_stream(
        session_uuid="testuuid"
    )
    chunks: List[LLMStreamResponse] = []
    for chunk in res:
        chunks.append(chunk)

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0


@pytest.mark.asyncio
async def test_chat_astream(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )

    res: AsyncGenerator[LLMStreamResponse, None] = proxy.chat_astream(
        session_uuid="testuuid"
    )
    chunks: List[LLMStreamResponse] = []
    async for chunk in res:
        chunks.append(chunk)

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert chunks[-1].api_response is not None
    assert len([chunk for chunk in chunks if chunk.error is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.parsed_outputs is not None]) == 0
    assert len([chunk for chunk in chunks if chunk.raw_output is not None]) > 0


def test_chat_run_extra_long_input(
    mocker,
    mock_fetch_chat_model: AsyncMock,
    mock_async_chat_log_to_cloud: AsyncMock,
):
    mocker.patch("promptmodel.llms.llm_proxy.get_max_tokens", return_value=10)

    fetch_chat_model = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy.fetch_chat_model", mock_fetch_chat_model
    )

    async_chat_log_to_cloud = mocker.patch(
        "promptmodel.llms.llm_proxy.LLMProxy._async_chat_log_to_cloud",
        mock_async_chat_log_to_cloud,
    )

    res: LLMResponse = proxy.chat_run(session_uuid="testuuid")

    fetch_chat_model.assert_called_once()
    async_chat_log_to_cloud.assert_called_once()

    assert res.raw_output is not None
    assert res.error is None or res.error is False
    assert res.api_response is not None
    assert res.parsed_outputs is None
    if isinstance(res.api_response.usage, dict):
        assert res.api_response.usage["prompt_tokens"] < 15
        print(res.api_response.usage["prompt_tokens"])
    else:
        assert res.api_response.usage.prompt_tokens < 15
        print(res.api_response.usage.prompt_tokens)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Coroutine, Union
from uuid import uuid4

from litellm import ModelResponse

from promptmodel import DevClient

from promptmodel.llms.llm_proxy import LLMProxy
from promptmodel.utils import logger
from promptmodel.utils.config_utils import (
    read_config,
    upsert_config,
    check_connection_status_decorator,
)
from promptmodel.utils.async_utils import run_async_in_sync
from promptmodel.types.response import LLMStreamResponse, LLMResponse, ChatModelConfig
from promptmodel.types.enums import InstanceType
from promptmodel.types.request import ChatLogRequest
from promptmodel.apis.base import AsyncAPIClient


class RegisteringMeta(type):
    def __call__(cls, *args, **kwargs):
        instance: ChatModel = super().__call__(*args, **kwargs)
        # Find the global client instance in the current context
        client = cls.find_client_instance()
        if client is not None:
            client.register_chat_model(instance.name)
        return instance

    @staticmethod
    def find_client_instance():
        import sys

        # Get the current frame
        frame = sys._getframe(2)
        # Get global variables in the current frame
        global_vars = frame.f_globals
        # Find an instance of Client among global variables
        for var_name, var_val in global_vars.items():
            if isinstance(var_val, DevClient):
                return var_val
        return None


class ChatModel(metaclass=RegisteringMeta):
    """

    Args:
        name (_type_): _description_
        version (Optional[ Union[str, int] ], optional): Choose which FunctionModel version to use. Defaults to "deploy". It can be "deploy", "latest", or version number.
        api_key (Optional[str], optional): API key for the LLM. Defaults to None. If None, use api_key in .env file.
    """

    def __init__(
        self,
        name,
        session_uuid: str = None,
        version: Optional[Union[str, int]] = "deploy",
        api_key: Optional[str] = None,
    ):
        self.name = name
        self.api_key = api_key
        self.llm_proxy = LLMProxy(name, version)
        self.version = version
        self.recent_log_uuid = None

        if session_uuid is None:
            self.session_uuid = str(uuid4())
            instruction, version_details, chat_logs = run_async_in_sync(
                LLMProxy.fetch_chat_model(self.name, None, version)
            )
            config = read_config()
            if (
                "connection" in config
                and "initializing" in config["connection"]
                and config["connection"]["initializing"] == True
            ):
                return
            elif (
                "connection" in config
                and "reloading" in config["connection"]
                and config["connection"]["reloading"] == True
            ):
                return
            else:
                run_async_in_sync(
                    self.llm_proxy._async_make_session_cloud(
                        self.session_uuid,
                        version_details["uuid"],
                    )
                )
        else:
            self.session_uuid = session_uuid

    @check_connection_status_decorator
    def get_config(
        self,
        *args,
        **kwargs,
    ) -> ChatModelConfig:
        """Get config for the ChatModel.
        It will fetch the published prompt and version config from the Cloud. (It will be saved in cache DB, so there is no extra latency for API call.)
        - If you made A/B testing in Web Dashboard, it will fetch the prompt randomly by the A/B testing ratio.
        If dev mode is initializing, it will return None

        Returns:
            ChatModelConfig: config for the ChatModel, which contains prompts and version_detail, message_logs
        """
        prompt, version_detail, message_logs = run_async_in_sync(
            LLMProxy.fetch_chat_model(self.name, self.session_uuid, self.version)
        )

        return ChatModelConfig(
            system_prompt=prompt,
            model=version_detail["model"],
            name=self.name,
            version_uuid=str(version_detail["uuid"]),
            version=version_detail["version"],
            message_logs=message_logs,
        )

    @check_connection_status_decorator
    def add_messages(
        self,
        new_messages: List[Dict[str, Any]],
        metadata_list: List[Optional[Dict]] = [],
        *args,
        **kwargs,
    ) -> None:
        """Add messages to the chat model.

        Args:
            new_messages (List[Dict[str, Any]]): list of messages. Each message is a dict with 'role', 'content', and 'function_call'.
        """
        # Save messages to Cloud DB
        log_uuid_list = [str(uuid4()) for _ in range(len(new_messages))]
        run_async_in_sync(
            self.llm_proxy._async_chat_log_to_cloud(
                session_uuid=str(self.session_uuid),
                version_uuid=None,
                chat_log_request_list=[
                    ChatLogRequest(**{"message": message, "uuid": str(uuid4())})
                    for message in new_messages
                ],
            )
        )

        self.recent_log_uuid = log_uuid_list[-1]

    @check_connection_status_decorator
    def run(
        self,
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        stream: Optional[bool] = False,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Run FunctionModel. It does not raise error.

        Args:
            functions (List[Dict[str, Any]], optional): list of functions to run. Defaults to None.

        Returns:
            LLMResponse: response from the promptmodel. you can find raw output in response.raw_output or response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        if stream:

            def gen():
                cache: Optional[LLMStreamResponse] = None

                for item in self.llm_proxy.chat_stream(
                    self.session_uuid, functions, tools
                ):
                    yield item
                    cache: LLMStreamResponse = item
                if cache:
                    self.recent_log_uuid = cache.pm_detail.log_uuid

            return gen()
        else:
            res = self.llm_proxy.chat_run(self.session_uuid, functions, tools)
            self.recent_log_uuid = res.pm_detail.log_uuid
            return res
        # return self.llm_proxy.chat_run(self.session_uuid, functions, self.api_key)

    @check_connection_status_decorator
    async def arun(
        self,
        functions: Optional[List[Dict[str, Any]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        stream: Optional[bool] = False,
        *args,
        **kwargs,
    ) -> LLMResponse:
        """Async run FunctionModel. It does not raise error.

        Args:
            functions (List[Dict[str, Any]], optional): list of functions to run. Defaults to None.

        Returns:
            LLMResponse: response from the promptmodel. you can find raw output in response.raw_output or response.api_response['choices'][0]['message']['content'].

        Error:
            It does not raise error. If error occurs, you can check error in response.error and error_log in response.error_log.
        """
        if stream:

            async def async_gen():
                cache: Optional[LLMStreamResponse] = None

                async for item in self.llm_proxy.chat_astream(
                    self.session_uuid, functions, tools
                ):
                    yield item
                    cache: LLMStreamResponse = item
                if cache:
                    self.recent_log_uuid = cache.pm_detail.log_uuid

            return async_gen()
        else:
            res: LLMResponse = await self.llm_proxy.chat_arun(
                self.session_uuid, functions
            )
            self.recent_log_uuid = res.pm_detail.log_uuid
            return res
        # return await self.llm_proxy.chat_arun(
        #     self.session_uuid, functions, self.api_key
        # )

    @check_connection_status_decorator
    async def log_score_to_session(
        self, score: Optional[Dict[str, Any]] = {}, *args, **kwargs
    ):
        try:
            res = await AsyncAPIClient.execute(
                method="POST",
                path="/save_chat_session_score",
                params={
                    "chat_session_uuid": self.session_uuid,
                },
                json=score,
                use_cli_key=False,
            )
            if res.status_code != 200:
                logger.error(f"Logging error: {res}")
        except Exception as exception:
            logger.error(f"Logging error: {exception}")

    @check_connection_status_decorator
    async def log_score(
        self,
        log_uuid: Optional[str] = None,
        score: Optional[Dict[str, Any]] = {},
        *args,
        **kwargs,
    ):
        """Log score for the ChatMessage in the Cloud DB.

        Args:
            log_uuid (Optional[str], optional): UUID of log to save score. Defaults to None. If None, it will use the recent log_uuid.
            score (Optional[Dict[str, Any]], optional): scores for message. Each keys will be created as Evaluation Metric in Promptmodel Dashboard. Defaults to {}.
        """
        try:
            if not log_uuid and self.recent_log_uuid:
                log_uuid = self.recent_log_uuid

            if log_uuid is None:
                raise ValueError(
                    "log_uuid is None. Please give log_uuid or run ChatModel.log or ChatModel.run first."
                )

            res = await AsyncAPIClient.execute(
                method="POST",
                path="/chat_message_score",
                params={
                    "chat_message_uuid": log_uuid,
                },
                json=score,
                use_cli_key=False,
            )
            if res.status_code != 200:
                logger.error(f"Logging error: {res}")
        except Exception as exception:
            logger.error(f"Logging error: {exception}")

    @check_connection_status_decorator
    async def log(
        self,
        openai_api_response: Optional[ModelResponse] = None,
        messages: Optional[List[Dict[str, Any]]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        *args,
        **kwargs,
    ) -> List[str]:
        """
        Log Messages to the Cloud DB.
        Args:
            openai_api_response (Optional[ModelResponse], optional): Response from the LLM. Defaults to None.
            messages (Optional[List[Dict[str, Any]]], optional): List of messages. Defaults to None.
            metadata (Optional[Dict[str, Any]], optional): Metadata for the log. Defaults to {}. It will be saved in the Last Log.

        You can give either openai_api_response or messages.
        If you give messages, it will log messages to the Cloud DB. This is useful when you want to log user messages to the Cloud DB.
        If you give openai_api_response, it will log messages to the Cloud DB. This is useful when you want to log the response of the LLM to the Cloud DB.

        Returns:
            List[str]: List of log_uuids in the Cloud DB. You can log scores for each logs using this log_uuid with method ChatModel.log_score.
        """
        try:
            if messages and openai_api_response:
                raise ValueError(
                    "You can give either openai_api_response or messages. You cannot give both."
                )

            chat_message_requests_body = []
            if messages:
                chat_message_requests_body = [
                    ChatLogRequest(
                        **{"message": message, "uuid": str(uuid4())}
                    ).model_dump()
                    for message in messages
                ]
            elif openai_api_response:
                chat_message_requests_body = [
                    ChatLogRequest(
                        **{
                            "message": openai_api_response.choices[
                                0
                            ].message.model_dump(),
                            "uuid": str(uuid4()),
                            "api_response": openai_api_response,
                        }
                    ).model_dump()
                ]

            chat_message_requests_body[-1]["metadata"] = metadata

            uuid_list = [x["uuid"] for x in chat_message_requests_body]

            res = await AsyncAPIClient.execute(
                method="POST",
                path="/chat_log",
                params={
                    "session_uuid": self.session_uuid,
                },
                json=chat_message_requests_body,
                use_cli_key=False,
            )
            if res.status_code != 200:
                logger.error(f"Logging error: {res}")

            # last log_uuid will be saved in the recent_log_uuid
            self.recent_log_uuid = uuid_list[-1]
            return uuid_list
        except Exception as exception:
            logger.error(f"Logging error: {exception}")
